# -*- coding: utf-8 -*-

import logging
from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance
from tingyun.config.settings import global_settings
from tingyun.logistics.object_name import callable_name
from tingyun.logistics.basic_wrapper import FunctionWrapper
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker

console = logging.getLogger(__name__)


def retrieve_tracker(headers, vendor='RabbitMQ'):
    """消费者调用这个接口，获取tracker，如果没有就自动生成一个并强制转换WebAction
    :return: tracker is_web
    """
    # 如果未webAction则跳过，只作为其附加组件
    tracker = current_tracker()
    if tracker:
        return tracker, True

    environ = headers
    if not headers:
        environ = {}

    tracker = Tracer(proxy_instance(), environ, vendor)
    tracker.start_work()

    if tracker.settings and not tracker.settings.mq.enabled:
        global_settings.enabled = False
        console.debug("Vendor %s consumer not enabled. Disable agent now", vendor)
        return None, False

    return tracker, False


def _do_comon_callback_wrap(wrapped, instance, args, kwargs):
    """
    :return:
    """
    tracker = current_tracker()
    if not tracker:
        console.debug("No tracker found for tracing %s", callable_name(wrapped))
        return wrapped(*args, **kwargs)

    with FunctionTracker(tracker, callable_name(wrapped)):
        return wrapped(*args, **kwargs)


def mq_common_callback_wrapper(wrapped, instance, args, kwargs):
    """
    :return:
    """
    def parse_callback(callback, *_args, **_kwargs):
        """
        :return:
        """
        return callback, _args, _kwargs

    cb, _args, _kwargs = parse_callback(*args, **kwargs)
    if getattr(cb, '_self_is_wrapped', None):
        console.debug("Consumer was wrapped before.")
        return wrapped(cb, *_args, **_kwargs)

    wrapped_cb = FunctionWrapper(cb, _do_comon_callback_wrap)
    wrapped_cb._self_is_wrapped = True

    return wrapped(wrapped_cb, *_args, **_kwargs)



# -*- coding: utf-8 -*-

"""define detector for tornado.gen
"""

import logging
from tingyun.armoury.ammunition.tornado_4.name_parser import object_name
from tingyun.armoury.ammunition.tornado_4.wrappers import wrap_function_wrapper, function_wrapper

from tingyun.armoury.ammunition.tornado_4.utils import obtain_current_tracker, NoneProxy
from tingyun.armoury.ammunition.function_tracker import FunctionTracker

console = logging.getLogger(__name__)

try:
    import sys
    get_frame = sys._getframe
except Exception, _:
    console.warning('You are using a python implementation without sys._getframe.')
    import inspect

    def _get_frame(depth):
        return inspect.stack(0)[depth]

    get_frame = _get_frame


def trace_runner_init(wrapped, instance, args, kwargs):
    """
    """
    tracker = obtain_current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    try:
        frame = get_frame(1)
    except ValueError:
        console.debug('tornado.gen.Runner is being created at the top of the stack. .')
        return wrapped(*args, **kwargs)

    max_frame_depth = 5
    frame_depth = 1
    while frame and frame_depth <= max_frame_depth:
        if frame.f_globals['__name__'] == 'tornado.gen':
            break
        frame = frame.f_back
        frame_depth += 1

    def _coroutine_name(func):
        return '%s %s' % (object_name(func), '(coroutine)')

    if '__name__' in frame.f_globals and frame.f_globals['__name__'] == 'tornado.gen' and \
       'func' in frame.f_locals and 'replace_callback' in frame.f_locals and frame.f_code.co_name == 'wrapper':

        instance._nb_coroutine_name = _coroutine_name(frame.f_locals['func'])

    tracker._ref_count += 1

    return wrapped(*args, **kwargs)


def trace_runner_run(wrapped, instance, args, kwargs):
    """
    """
    result = wrapped(*args, **kwargs)

    tracker = obtain_current_tracker()
    if tracker is None:
        return result

    if result is None:
        result = NoneProxy()

    if hasattr(instance, '_nb_coroutine_name') and instance._nb_coroutine_name is not None:
        result._nb_coroutine_name = instance._nb_coroutine_name

    if instance.finished:
        tracker._ref_count -= 1

    return result


@function_wrapper
def _do_wrap_decorated(wrapped, instance, args, kwargs):
    """
    """
    tracker = obtain_current_tracker()

    if tracker is None:
        return wrapped(*args, **kwargs)

    with FunctionTracker(tracker, name=object_name(wrapped)):
        return wrapped(*args, **kwargs)


def trace_coroutine(wrapped, instance, args, kwargs):
    """
    """
    func = wrapped(*args, **kwargs)
    return _do_wrap_decorated(func)


def trace_engine(wrapped, instance, args, kwargs):
    """
    """
    func = wrapped(*args, **kwargs)
    return _do_wrap_decorated(func)


def detect_gen(module):
    """
    """
    wrap_function_wrapper(module, 'Runner.__init__', trace_runner_init)
    wrap_function_wrapper(module, 'Runner.run', trace_runner_run)
    wrap_function_wrapper(module, 'coroutine', trace_coroutine)
    wrap_function_wrapper(module, 'engine', trace_engine)




"""this module is implement the function detector for webpy

"""
import logging
import sys
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.packages import six

console = logging.getLogger(__name__)


def app_delegate_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, instance, args, kwargs):
        f, fvars, fargs = args
        tracker = current_tracker()
        if not tracker:
            return wrapped(f, fvars, fargs)

        if isinstance(f, six.string_types):
            name = f
        else:
            name = callable_name(f)

        tracker.set_tracker_name(name, 2)
        with FunctionTracker(tracker, name):
            try:
                return wrapped(*args, **kwargs)
            except Exception:
                tracker.record_exception(*sys.exc_info())
                raise

    return FunctionWrapper(wrapped, wrapper)


def trace_app_views(module, object_path):
    wrap_object(module, object_path, app_delegate_wrapper)




"""

"""

import sys
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, process_cross_trace, parse_parameters

from tingyun.armoury.ammunition.external_tracker import MALFORMED_URL_ERROR_CODE, UNKNOWN_HOST_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CONNECT_EXCEPTION_ERROR_CODE, TIMEOUT_EXCEPTION_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CLIENT_PROTOCOL_ERROR_CODE, ILLEGAL_RESPONSE_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE, SSL_EXCEPTION_ERROR_CODE


try:
    from urllib3.exceptions import TimeoutError
except ImportError:
    class TimeoutError(Exception):
        pass

try:
    from urllib3.exceptions import SSLError
except ImportError:
    class SSLError(Exception):
        pass

try:
    from urllib3.exceptions import DecodeError
except ImportError:
    class DecodeError(Exception):
        pass

try:
    from urllib3.exceptions import ResponseError
except ImportError:
    class ResponseError(Exception):
        pass


try:
    from urllib3.exceptions import ReadTimeoutError
except ImportError:
    class ReadTimeoutError(Exception):
        pass

try:
    from urllib3.exceptions import ConnectTimeoutError
except ImportError:
    class ConnectTimeoutError(Exception):
        pass

try:
    from urllib3.exceptions import ProtocolError
except ImportError:
    class ProtocolError(Exception):
        pass


try:
    from urllib3.exceptions import ResponseNotChunked
except ImportError:
    class ResponseNotChunked(Exception):
        pass

try:
    from urllib3.exceptions import HostChangedError
except ImportError:
    class HostChangedError(Exception):
        pass

try:
    from urllib3.exceptions import LocationParseError
except ImportError:
    class LocationParseError(Exception):
        pass

try:
    from urllib3.exceptions import LocationValueError
except ImportError:
    class LocationValueError(Exception):
        pass


try:
    from urllib3.exceptions import ProxyError
except ImportError:
    class ProxyError(Exception):
        pass


def wrap_exception(wrapped, action_url, parameters, *args, **kwargs):
    """
    """
    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = action_url if not callable(action_url) else action_url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)

    try:
        rtv = wrapped(*args, **kwargs)
    except SSLError:
        error_code = SSL_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except (DecodeError, ResponseError, ResponseNotChunked):
        error_code = ILLEGAL_RESPONSE_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except (ConnectTimeoutError, ProxyError):
        error_code = CONNECT_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except ProtocolError:
        error_code = CLIENT_PROTOCOL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except (HostChangedError, LocationParseError):
        error_code = UNKNOWN_HOST_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except LocationValueError:
        error_code = MALFORMED_URL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except (TimeoutError, ReadTimeoutError):
        error_code = TIMEOUT_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise
    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='urllib3')
        raise

    if int(getattr(rtv, 'status', 0)) != 200:
        tracker.record_external_error(url, error_code, getattr(rtv, 'status', 500), sys.exc_info(), params,
                                      module_name='urllib3')

    return rtv


def detect(module):
    """
    """
    def url_url_open(instance, method, url, *args, **kwargs):
        return url

    def parse_params(method, url, redirect=True, **kw):
        _args = (method, url)
        _kwargs = kw
        protocol = 'http' if 'https' not in url else 'https'

        if not kw:
            _kwargs = {}

        ty_headers, external_id = process_cross_trace(_kwargs.get("headers", None), protocol)
        _kwargs["redirect"] = redirect
        _kwargs["headers"] = ty_headers

        return external_id, _args, _kwargs

    wrap_external_trace(module, 'PoolManager.urlopen', 'urllib3', url_url_open, params=parse_params,
                        exception_wrapper=wrap_exception)



# -*- coding: utf-8 -*-

"""
"""

import traceback
import logging

from tingyun.armoury.ammunition.timer import Timer
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.warehouse.database_node import DatabaseNode

console = logging.getLogger(__name__)


class DatabaseTracker(Timer):
    """
    """
    def __init__(self, tracker, sql, dbtype='', dbapi=None, connect_params=None, cursor_params=None,
                 execute_params=None, host="Unknown", port="Unknown", db_name="Unknown"):
        """
        :param tracker:
        :param sql:
        :param dbapi:
        :param connect_params:
        :param cursor_params:
        :param execute_params:
        :return:
        """
        super(DatabaseTracker, self).__init__(tracker)

        self.dbapi = dbapi
        self.connect_params = connect_params
        self.cursor_params = cursor_params
        self.execute_params = execute_params
        self.sql = sql
        self.db_type = dbtype
        self.host = host
        self.port = port
        self.db_name = db_name

        self.stack_trace = None
        self.sql_format = None

    def finalize_data(self):
        """create all the data if need
        :return:
        """
        connect_params = None
        cursor_params = None
        execute_params = None
        settings = self.tracker.settings

        if settings.action_tracer.enabled and self.duration >= settings.action_tracer.stack_trace_threshold:
            if self.tracker.stack_trace_count < settings.stack_trace_count:
                self.stack_trace = traceback.extract_stack()
                self.tracker.stack_trace_count += 1

        explain_enabled = settings.action_tracer.explain_enabled
        explain_threshold = settings.action_tracer.explain_threshold
        if settings.action_tracer.enabled and explain_enabled and self.duration > explain_threshold:
            if self.tracker.explain_plan_count < settings.explain_plan_count:
                connect_params = self.connect_params
                cursor_params = self.cursor_params
                execute_params = self.execute_params

        self.sql_format = settings.action_tracer.record_sql
        self.connect_params = connect_params
        self.cursor_params = cursor_params
        self.execute_params = execute_params

    def create_node(self):
        tracker = current_tracker()
        if tracker:
            tracker.db_time = self.duration

        return DatabaseNode(dbapi=self.dbapi, sql=self.sql, children=self.children, start_time=self.start_time,
                            end_time=self.end_time, duration=self.duration, exclusive=self.exclusive,
                            dbtype=self.db_type, stack_trace=self.stack_trace, sql_format=self.sql_format,
                            connect_params=self.connect_params, cursor_params=self.cursor_params,
                            execute_params=self.execute_params,  host=self.host, port=self.port, db_name=self.db_name)

    def terminal_node(self):
        return True



import copy
import logging
import os
import json
import re

from tingyun import get_version

console = logging.getLogger(__name__)
filter_pro = ["log_level_mapping", ]


class Settings(object):
    """
    """
    def __repr__(self):
        return repr(self.__dict__)

    def __iter__(self):
        """return the iteration for setting item
        """
        return iter(flatten_settings(self).items())


class ErrorCollectorSettings(Settings):
    pass


class ActionTracerSettings(Settings):
    pass


class Naming(Settings):
    """naming Rules class for use.
    """
    def __init__(self, rules=None):
        self.rules = json.loads(rules) if not isinstance(rules, list) else rules
        self.cached = {}  # {url:metric}
        self.regex_url = {}
        self.loaded = False
        self.method = {"0": "GET|POST|PUT|DELETE|HEAD", "1": "GET", "2": "POST", "3": "PUT", "4": "DELETE", "5": "HEAD"}
        self.match_type = {"1": self.match_equal, "2": self.match_start_with, "3": self.match_end_with,
                           "4": self.match_include, "5": self.match_regex, "0": self.any_value}

    @property
    def should_naming(self):
        return True if self.rules else False

    def any_value(self, src, dist, flag=0):
        return True

    def match_equal(self, src, dist, flag=0):
        """
        :param src: compare source data
        :param dist: compare target
        :return:
        """
        if isinstance(src, list):
            return dist in src

        return src == dist

    def match_start_with(self, src, dist, flag=0):
        if isinstance(src, list):
            for s in src:
                if str(s).startswith(dist):
                    return True
        else:
            return str(src).startswith(dist)

        return False

    def match_end_with(self, src, dist, flag=0):
        if isinstance(src, list):
            for s in src:
                if str(s).endswith(dist):
                    return True
        else:
            return str(src).endswith(dist)

        return False

    def match_include(self, src, dist, flag=0):
        if isinstance(src, list):
            for s in src:
                if str(dist) in str(s):
                    return True
        else:
            return str(dist) in str(src)

        return False

    def match_regex(self, src, dist, flag=re.IGNORECASE):

        if dist in self.regex_url:
            return self.regex_url.get(dist)

        def _match_value(_src, _dist, cache):
            for s in _src:
                pattern = re.compile(_dist, flag)
                match = pattern.match(s)

                if match:
                    cache[dist] = pattern
                    return True

            return False

        if isinstance(src, list):
            return _match_value(src, dist, self.regex_url)
        else:
            return _match_value([src], dist, self.regex_url)

    def naming_url_parameters(self, rule, params, grouped_param):
        """
        :param params:dict
        :param rule: rule
        :return:
        """
        if not rule:
            return ""
        p = "&".join(["%s=%s" % (k, params.get(k)) for k in rule.split(",") if k in params])
        grouped_param.append(p)
        return p

    def naming_web_action(self, method, uri, url_param, header_param, cookie_param):
        """ entrance for naming web action with rules
        :param cookie_param:
        :param url_param:
        :param header_param:
        :param method: current http request method to the Application
        :param uri: uri of current request
        :return:
        """
        if uri in self.cached:
            return self.cached.get(uri)

        if self.rules and not self.loaded:
            self.rules = json.loads(self.rules) if not isinstance(self.rules, list) else self.rules
            self.loaded = True

        if not self.rules:
            return None

        if not ("name" in self.rules[0] and "match" in self.rules[0] and "split" in self.rules[0]):
            self.rules = None
            return None

        for rule in self.rules:
            # match the request method
            if method not in self.method.get(str(rule["match"]["method"]), ""):
                continue

            # match the url rule, url case insensitive
            match_func = self.match_type[str(rule["match"]["match"])]
            if not match_func(uri, rule["match"]["value"], re.IGNORECASE):
                continue

            # match the parameters rule,not support body parameters, when match the value, the case sensitive
            # And the relationship between parameters and url matching is logic 'and'
            matched_param = True
            param_type = {"1": url_param, "2": header_param, "3": {}, "4": cookie_param}
            for params in rule["match"]['params']:
                assert isinstance(params, dict)

                src_value = param_type.get(str(params['type'])).get(params["name"], "")
                match_func = self.match_type[str(params["match"])]

                if not match_func(src_value, params["value"]):
                    matched_param = False
                    break

            # url match & param should be matched
            if not matched_param:
                continue

            named_metric = ["/", rule["name"], "/"]
            uri_rule = rule["split"]["uri"]
            if uri_rule:
                uri_section = uri.strip('/').split("/")
                uri_rules = [ur for ur in uri_rule.split(",") if ur]

                if 1 == len(uri_rules) and 0 < int(uri_rules[0]):
                    named_metric.append("/".join(uri_section[: int(uri_rules[0])]))
                elif 1 == len(uri_rules) and 0 > int(uri_rules[0]):
                    named_metric.append("/".join(uri_section[int(uri_rules[0]):]))
                elif len(uri_rules) > 1:
                    tmp = [uri_section[int(ux) - 1] for ux in uri_rules if int(ux) <= len(uri_section)]
                    named_metric.append("/".join(tmp))

            named_metric.append("?")
            grouped_param = []
            self.naming_url_parameters(rule["split"]["urlParams"], url_param, grouped_param)
            self.naming_url_parameters(rule["split"]["headerParams"], header_param, grouped_param)
            self.naming_url_parameters(rule["split"]["cookieParams"], cookie_param, grouped_param)
            named_metric.append("&".join(grouped_param))

            if rule['split']['method']:
                named_metric.append("(%s)" % method)

            return "".join(named_metric)

        return None


class TransactionTracerSettings(Settings):
    pass


class RumTraceSettings(Settings):
    pass


class MQ(Settings):
    pass


_settings = Settings()
_settings.transaction_tracer = TransactionTracerSettings()
_settings.error_collector = ErrorCollectorSettings()
_settings.action_tracer = ActionTracerSettings()
_settings.rum = RumTraceSettings()
_settings.naming = Naming([])
_settings.mq = MQ

# configure file
_settings.x_tingyun_id = "%s;c=1;x=%s;e=%s;s=%s"
_settings.app_name = "Python App"
_settings.plugins = ''  # split with ','
_settings.license_key = "This is default license"
_settings.enabled = True
_settings.log_file = None
_settings.log_level = logging.INFO
_settings.audit_mode = False
_settings.ssl = True
_settings.daemon_debug = False
_settings.host = "redirect.networkbench.com"
_settings.port = ''
_settings.tingyunIdSecret = ''
_settings.dataSentInterval = 60
_settings.applicationId = 'do-not-get-application-id'
_settings.appSessionKey = "do-not-get-session-key"
_settings.urls_merge = True
_settings.verify_certification = True
_settings.tornado_wsgi_adapter_mode = False

# for proxy
_settings.proxy_host = ''
_settings.proxy_port = ''
_settings.proxy_user = ''
_settings.proxy_pwd = ''
_settings.proxy_scheme = 'http'

# python environment variable
_settings.config_file = None
_settings.enable_profile = True
_settings.max_profile_depth = 600

# cross transaction trace
_settings.transaction_tracer.enabled = False

# internal use constance
_settings.shutdown_timeout = float(os.environ.get("TINGYUN_AGENT_SHUTDOWN_TIMEOUT", 2.5))
_settings.startup_timeout = float(os.environ.get("TINGYUN_AGENT_STARTUP_TIMEOUT", 0.0))
_settings.data_version = '1.2.0'
_settings.agent_version = get_version()
_settings.data_report_timeout = 15.0
_settings.stack_trace_count = 30  # used to limit the depth of action tracer stack
_settings.explain_plan_count = 30  # used to limit the depth of sql explain tracer stack
_settings.action_tracer_nodes = 2000
_settings.slow_sql_count = 20
_settings.action_apdex = {}
_settings.web_action_uri_params_captured = {}
_settings.external_url_params_captured = {}

# server configuration & some limitation about the data
# set the default value for it.
_settings.action_tracer.enabled = True
_settings.action_tracer.action_threshold = 2 * 1000  # 2s
_settings.action_tracer.top_n = 48
_settings.action_tracer.stack_trace_threshold = 500  # 500ms

_settings.action_tracer.slow_sql = True
_settings.action_tracer.slow_sql_threshold = 500
_settings.action_tracer.log_sql = False
_settings.action_tracer.explain_enabled = True
_settings.action_tracer.explain_threshold = 500  # 500ms
_settings.action_tracer.record_sql = "obfuscated"

# rum tracer settings default value
_settings.rum.enabled = False
_settings.rum.script = None
_settings.rum.sample_ratio = 1.0

_settings.auto_action_naming = True
_settings.urls_captured = []
_settings.ignored_params = []

_settings.error_collector.enabled = True
_settings.error_collector.ignored_status_codes = []

_settings.apdex_t = 500
_settings.capture_params = False
_settings.quantile = []
_settings.quantile_org = []
_settings.min_quantile_length = 4


# mq consumer
_settings.mq.enabled = False


_settings.log_level_mapping = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
}


def flatten_settings(settings):
    """get the iteration for setting item, include Settings object in Setting
    """

    def _flatten(settings_item, name, setting_object):
        for key, value in setting_object.__dict__.items():
            if key in filter_pro:
                console.debug("skip flatten property %s", key)
                continue

            if isinstance(value, Settings):
                if name:
                    _flatten(settings_item, '%s.%s' % (name, key), value)
                else:
                    _flatten(settings_item, key, value)
            else:
                if name:
                    settings_item['%s.%s' % (name, key)] = value
                else:
                    settings_item[key] = value

        return settings_item

    return _flatten({}, None, settings)


def get_upload_settings():
    """ construct the settings for upload to server
    :return: settings
    """
    dump_settings = {
        "nbs.license_key": _settings.license_key,
        "nbs.agent_enabled": _settings.enabled,
        "nbs.app_name": _settings.app_name,
        "nbs.auto_app_naming": False,
        "nbs.agent_log_file_name": _settings.log_file,
        "nbs.audit_mode": _settings.audit_mode,
        "nbs.agent_log_level": _settings.log_level,
        "nbs.proxy_host": _settings.proxy_host,
        "nbs.proxy_port": _settings.proxy_port,
        "nbs.proxy_user": _settings.proxy_user,
        "nbs.proxy_password": _settings.proxy_pwd,
        "nbs.host": _settings.host,
        "nbs.port": _settings.port,
        "nbs.ssl": _settings.ssl,
        "nbs.action_tracer.log_sql": _settings.action_tracer.log_sql,
        "nbs.daemon_debug": _settings.daemon_debug,
    }

    return dump_settings


def apply_config_setting(settings_object, name, value):
    target = settings_object
    fields = name.split('.', 1)

    while len(fields) > 1:
        # for additional Settings from server
        if not hasattr(target, fields[0]):
            setattr(target, fields[0], Settings())
        target = getattr(target, fields[0])
        fields = fields[1].split('.', 1)

    # transfer the string ignored ignored_status_codes to list
    if fields[0] == "ignored_status_codes" and value:
        try:
            value = [int(v) for v in value.strip().split(",")]
        except Exception as err:
            console.warning("got invalid ignore status code %s, errors %s", value, err)
    elif fields[0] == "ignored_status_codes" and not value:
        value = []

    # transfer the urls_captured to list
    # Warning in windows, the split with \n will get incorrect result
    if fields[0] == "urls_captured" and value:
        try:
            value = [re.compile(r"%s" % v) for v in value.strip().split("\n")]
        except Exception as err:
            console.warning("compile re url failed, %s, %s", value, err)
    elif fields[0] == "urls_captured" and not value:
        value = []

    # transfer the ignored_params to list
    if fields[0] == "ignored_params" and value:
        value = [str(v) for v in value.strip().split(",")]
    elif fields[0] == "ignored_params" and not value:
        value = []

    setattr(target, fields[0], value)


def uri_params_captured(original):
    """parse the web_action_uri_params_captured with protocal
    :param original:
    :return: formatted uri with param
    """
    uri_params = {}
    try:
        for uri_param in original.split('|'):
            parts = uri_param.split(",")
            uri = parts.pop(0).encode('utf-8')
            if not uri:
                continue

            uri_params[uri] = str(parts)
    except Exception as err:
        console.error("Parse the uri occurred errors. %s", err)

    return uri_params


def merge_settings(server_side_config=None):
    """
    :param server_side_config: the config downloaded from server
    :return: the merged settings
    """
    console.info("get the server side config %s", server_side_config)
    settings_snapshot = copy.deepcopy(_settings)
    global_conf_name = ["applicationId", "enabled", "appSessionKey", "dataSentInterval", "apdex_t", "tingyunIdSecret"]

    if not server_side_config:
        console.warning("update server config failed %s, local settings will be used.", server_side_config)
        return settings_snapshot

    # pop the structured config
    server_conf = server_side_config.pop("config", {})

    # get the individual settings
    if 'actionApdex' in server_side_config:
        original_apdex = server_side_config.pop("actionApdex", {})
        urls_apdexs = dict((key, original_apdex[key]) for key in original_apdex)
        action_apdex_urls = dict((key.encode("utf8"), urls_apdexs[key]) for key in urls_apdexs if key)

        server_side_config.update({"action_apdex": action_apdex_urls})

    for name in global_conf_name:
        if name not in server_side_config:
            console.warning("Lost server configure %s", name)
            continue

        apply_config_setting(settings_snapshot, name, server_side_config.get(name, ""))

    # transfer specified params captured with same uri
    if 'nbs.web_action_uri_params_captured' in server_conf:
        web_uri_params_captured = server_conf.pop("nbs.web_action_uri_params_captured", '')
        settings_snapshot.web_action_uri_params_captured = uri_params_captured(web_uri_params_captured)

    if 'nbs.external_url_params_captured' in server_conf:
        external_uri_params_captured = server_conf.pop("nbs.external_url_params_captured", '')
        settings_snapshot.external_url_params_captured = uri_params_captured(external_uri_params_captured)

    if 'nbs.quantile' in server_conf:
        quantile = server_conf.pop("nbs.quantile", '[]')
        try:
            quantile = json.loads(quantile)
        except Exception as err:
            console.error("Error type of quantile value %s, error %s", quantile, err)
        else:
            if len(quantile) != len(set(quantile)):
                console.warning("Error quantile value with duplicated %s", quantile)
            else:
                tmp = [v * 1.0 for v in quantile if str(v).isdigit()]
                if len(tmp) != len(quantile):
                    console.warning("quantile value is not digit.")
                else:
                    settings_snapshot.quantile_org = list(quantile)
                    settings_snapshot.quantile = tmp

    for item in server_conf:
        value = server_conf[item]
        # drop the first part of name with point
        start_pos = str(item).find('.') + 1
        apply_config_setting(settings_snapshot, item[start_pos:], value)

    console.info("return merged settings %s", settings_snapshot)
    return settings_snapshot


def global_settings():
    return _settings




"""define the transportation engine for upload the data.

"""
import time
import logging
import socket
from tingyun.config.settings import global_settings, merge_settings, get_upload_settings
from tingyun.logistics.exceptions import RetryDataForRequest
from tingyun.packages import requests
from tingyun.logistics.transportation import transmitter

console = logging.getLogger(__name__)


class Engine(object):
    """this class hold the application session information from headquarters
    """
    def __init__(self, url, config, license_key, redirect_host):
        self.url = url
        self.config = config
        self.license_key = license_key
        self.local_setting = global_settings()
        self.merged_settings = config
        self.redirect_host = redirect_host

        self._session = None
        self.request_param = {
            "licenseKey": self.license_key,
            "version": self.local_setting.data_version,
            "appSessionKey": self.config.appSessionKey,
        }

    @property
    def session(self):
        """
        :return:
        """
        if self._session is None:
            self._session = requests.session()

        return self._session

    def close_session(self):
        """use to shutdown session, report status to server for test.
        """
        self._session.close()
        self._session = None

    def send_performance_metric(self, metric_data):
        """
        """
        return transmitter(self.session, self.url, "send_collect_data", metric_data, self.request_param,
                           self.config.audit_mode)

    def send_error_trace(self, error_trace):
        """
        """
        if not error_trace:
            return

        return transmitter(self.session, self.url, "send_error_trace", error_trace, self.request_param,
                           self.config.audit_mode)

    def send_action_trace(self, action_trace):
        """
        """
        if not action_trace:
            return

        return transmitter(self.session, self.url, "send_action_trace", action_trace, self.request_param,
                           self.config.audit_mode)

    def send_sql_trace(self, sql_trace):
        """
        """
        if not sql_trace:
            return

        return transmitter(self.session, self.url, "send_sql_trace", sql_trace, self.request_param,
                           self.config.audit_mode)

    def send_profile_data(self, data):
        """it allow to send empty data.
        """
        audit_mode = self.config.audit_mode
        url = connect_url("uploadAgentCommands", self.redirect_host)
        if not data:
            return

        return transmitter(self.session, url, "send_profile_data", data, self.request_param, audit_mode)

    def request_agent_commands(self):
        """
        """
        url = connect_url("getAgentCommands", self.redirect_host)
        audit_mode = self.config.audit_mode
        result = transmitter(self.session, url, "get_agent_commands", {}, self.request_param, audit_mode)

        return result

    def send_external_error_trace(self, external_error_trace):
        """
        """
        if not external_error_trace:
            return

        return transmitter(self.session, self.url, "send_external_error_trace", external_error_trace,
                           self.request_param, self.config.audit_mode)


def connect_url(action, host=None, with_port=False):
    """
    :param action: the uploader actually intention
    :return: ruled url, the suitable url according to headquarters strategy
    """
    settings = global_settings()
    scheme = settings.ssl and 'https' or 'http'
    url_reg = "%s://%s/%s"

    if host is None:
        host = settings.host

    url = url_reg % (scheme, host, action)
    if with_port and settings.port:
        url_reg = "%s://%s:%s/%s"
        url = url_reg % (scheme, host, settings.port, action)

    return url


def create_connection(license_key, app_name, linked_applications, machine_env, settings):
    """
    :return:
    """
    start_time = time.time()

    if not license_key:
        license_key = settings.license_key

    if not license_key:
        console.error("license key should be provided in tingyun.ini config file, agent may can not started.")

    try:
        console.info("init app with license: %s, app: %s, link app: %s, env: %s, settings: %s, upload settings: %s",
                     license_key, app_name, linked_applications, machine_env, settings, get_upload_settings())

        # when private the agent. we should get redirect with port set in settings
        url = connect_url("getRedirectHost", with_port=True)
        param = {"licenseKey": license_key, "version": settings.data_version}
        redirect_host = transmitter(None, url, "getRedirectHost", {}, param)
        local_conf = {
            "host": socket.gethostname(),
            "appName": [app_name] + linked_applications,
            "language": "Python",
            "agentVersion": settings.agent_version,
            "config": get_upload_settings(),
            "env": machine_env
        }

        # At second step. we need not the port with url. because the returned redirect url take it or has default 80.
        url = connect_url("initAgentApp", redirect_host, with_port=False)
        server_conf = transmitter(None, url, "initAgentApp", local_conf, {"licenseKey": license_key})
        app_config = merge_settings(server_conf)
    except Exception as err:
        console.error("errors when connect to server %s", err)
        raise RetryDataForRequest("Errors occurred, when create session with server. %s" % err)
    else:
        console.info("Successful register application to agent server with name: %s, use time: %ss", app_name,
                     time.time() - start_time)

        url = connect_url("upload", redirect_host)
        return Engine(url, app_config, license_key, redirect_host)




"""Define this module for basic armory for detect jinja2

"""

from tingyun.armoury.ammunition.function_tracker import wrap_function_trace


def detect_template_loader(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, 'BaseLoader.load'):
        wrap_function_trace(module, 'BaseLoader.load')


def detect_jinja2(module):
    """
    """
    def parse_tempate_name(instance, *args, **kwargs):
        return instance.name or instance.filename

    wrap_function_trace(module, 'Template.render', parse_tempate_name, 'Template.Render')

    def parse_template_for_env(instance, source, name=None, *args, **kwargs):
        """
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        return name or 'template'

    wrap_function_trace(module, 'Environment.compile', parse_template_for_env, 'Template.compile')




"""this module defined to get the cpu usage/Utilization
"""

import os
import logging
from tingyun.logistics.attribution import TimeMetric
from tingyun.armoury.sampler.system_info import cpu_count

console = logging.getLogger(__name__)


class CpuUsage(object):
    """
    """
    def __init__(self, *args):
        """
        :return:
        """
        self.__times = None
        self.__cpu_count = 0

    def start(self):
        """
        :return:
        """
        try:
            self.__times = os.times()
        except Exception as _:
            self.__times = None

        self.__cpu_count = cpu_count()

    def stop(self):
        """
        :return:
        """
        self.__times = None

    def __call__(self, *args, **kwargs):
        """
        :param args:
        :param kwargs:
        :return:
        """
        duration = 60
        if len(args) != 0:
            duration = args[0]

        if self.__times is None:
            return

        now = os.times()
        cpu_use_time = (now[0] + now[1]) - (self.__times[0] + self.__times[1])
        self.__times = now

        cpu_use_metric = "GENERAL/CPU/NULL/UserTime"
        yield [TimeMetric(name=cpu_use_metric, scope=cpu_use_metric, duration=int(cpu_use_time),
                          exclusive=int(cpu_use_time)), ]

        if not self.__cpu_count:
            console.debug("Cpu count is 0, skip count the cpu utilization")
            return

        cpu_utilization_metric = 'GENERAL/CPU/NULL/UserUtilization'
        utilization = int((cpu_use_time / duration / self.__cpu_count) * 100)
        yield [TimeMetric(name=cpu_utilization_metric, scope=cpu_utilization_metric,
                          duration=utilization, exclusive=utilization), ]


cpu_usage_sampler = CpuUsage



# -*- coding: utf-8 -*-

import sys

from tingyun.armoury.ammunition.tornado.util import record_exception
from tingyun.armoury.ammunition.tornado.util import TrackerTransferContext
from tingyun.armoury.ammunition.tornado.wrappers import wrap_function_wrapper


def trace_handle_callback_exception(wrapped, instance, args, kwargs):
    """
    """
    record_exception(sys.exc_info())

    return wrapped(*args, **kwargs)


def trace_call_at(wrapped, instance, args, kwargs):
    """
    """
    with TrackerTransferContext(None):
        return wrapped(*args, **kwargs)


def trace_add_handler(wrapped, instance, args, kwargs):
    """
    """
    with TrackerTransferContext(None):
        return wrapped(*args, **kwargs)


def detect_ioloop(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, 'IOLoop.handle_callback_exception', trace_handle_callback_exception)
    wrap_function_wrapper(module, 'PollIOLoop.call_at', trace_call_at)
    wrap_function_wrapper(module, 'PollIOLoop.add_handler', trace_add_handler)




from collections import namedtuple
from tingyun.logistics.attribution import TimeMetric, node_start_time, node_end_time


_RedisNode = namedtuple('_RedisNode', ['command', 'children', 'start_time', 'end_time', 'duration', 'exclusive',
                                       'host', 'port', 'db'])


class RedisNode(_RedisNode):
    """
    """
    def time_metrics(self, root, parent):
        """
        :param root:
        :param parent:
        :return:
        """
        command = str(self.command).upper()

        name = 'GENERAL/Redis/%s:%s/All' % (self.host, self.port)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/Redis/NULL/All'
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebAction':
            name = "GENERAL/Redis/NULL/AllWeb"
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)
        else:
            name = "GENERAL/Redis/NULL/AllBackgound"
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'Redis/%s:%s%%2F%s/%s' % (self.host, self.port, self.db, command)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/Redis/%s:%s%%2F%s/%s' % (self.host, self.port, self.db, command)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, root):
        """
        :param root:
        :return:
        """
        command = str(self.command).upper()
        params = {}
        children = []
        call_count = 1
        class_name = ""
        method_name = root.name
        root.trace_node_count += 1
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        metric_name = 'Redis/%s:%s%%2F%s/%s' % (self.host, self.port, self.db, command)
        call_url = metric_name

        return [start_time, end_time, metric_name, call_url, call_count, class_name, command, params, children]



# -*- coding: utf-8 -*-

"""该模块用于解析对象的名称
"""

import sys
import types
import inspect
import functools

from tingyun.packages import six


def parse_module_name(obj):
    module_name = None

    if hasattr(obj, '__objclass__'):
        module_name = getattr(obj.__objclass__, '__module__', None)

    if module_name is None:
        module_name = getattr(obj, '__module__', None)

    if module_name is None:
        self = getattr(obj, '__self__', None)
        if self is not None and hasattr(self, '__class__'):
            module_name = getattr(self.__class__, '__module__', None)

    if module_name is None and hasattr(obj, '__class__'):
        module_name = getattr(obj.__class__, '__module__', None)

    if module_name and module_name not in sys.modules:
        module_name = '<%s>' % module_name

    if not module_name:
        module_name = '<unknown>'

    return module_name


def _object_context_py2(obj):
    class_name = None
    func_name = None

    if inspect.isclass(obj) or isinstance(obj, type):
        class_name = obj.__name__

    elif inspect.ismethod(obj):
        if obj.im_self is not None:
            class_name = getattr(obj.im_self, '__name__', None)
            if class_name is None:
                class_name = getattr(obj.im_self.__class__, '__name__')

        else:
            class_name = obj.im_class.__name__

        func_name = obj.__name__

    elif inspect.isfunction(obj):
        func_name = obj.__name__

    elif inspect.isbuiltin(obj):
        if obj.__self__ is not None:
            class_name = getattr(obj.__self__, '__name__', None) or getattr(obj.__self__.__class__, '__name__')

        func_name = obj.__name__

    elif isinstance(obj, types.InstanceType):
        func_name = getattr(obj, '__name__', None)

        if func_name is None:
            class_name = obj.__class__.__name__

    elif hasattr(obj, '__class__'):
        func_name = getattr(obj, '__name__', None)

        if func_name:
            if hasattr(obj, '__objclass__'):
                class_name = obj.__objclass__.__name__
            elif not hasattr(obj, '__get__'):
                class_name = obj.__class__.__name__
        else:
            class_name = obj.__class__.__name__

    path = class_name if class_name else ''
    if func_name:
        if path:
            path += '.'
        path += func_name

    owner = ''
    if inspect.ismethod(obj):
        if obj.__self__ is not None:
            class_name = getattr(obj.__self__, '__name__', None)
            if class_name is None:
                owner = obj.__self__.__class__  # bound method
            else:
                owner = obj.__self__  # class method
        else:
            owner = getattr(obj, 'im_class', None)  # unbound method

    module_name = parse_module_name(owner or obj)

    return module_name, path


def object_context(target):
    """
    :param target: 目标对象
    :return: (module, object_path)
    """

    # 对象被 functools.partial 打包
    if isinstance(target, functools.partial):
        target = target.func

    target_name = getattr(target, '_nb_object_name', None)
    if target_name:
        return target_name

    org_object = getattr(target, '_previous_object', None)
    if org_object:
        target_name = getattr(org_object, '_nb_object_name', None)

        if target_name:
            return target_name
    else:
        org_object = target

    # 没有被缓存，重新计算, 暂未考虑py3
    target_name = _object_context_py2(org_object)

    try:
        # 缓存当前名称
        if target is not org_object:
            target._nb_object_name = target_name

        org_object._nb_object_name = target_name
    except Exception as _:
        pass

    return target_name


def object_name(obj, separator=':'):
    """
    :param obj:
    :param separator:
    :return: module:object_path
    """
    return separator.join(object_context(obj))



# -*- coding: utf-8 -*-

"""communication proxy to dispatcher

"""
import logging
import threading
from tingyun.battlefield.dispatcher import dispatcher_instance
from tingyun.config.settings import global_settings

console = logging.getLogger(__name__)


class Proxy(object):
    """
    """
    _lock = threading.Lock()
    _instances = {}

    def __init__(self, dispatcher, app_name, link_app_name=None):
        """
        :param dispatcher:
        :param app_name:
        :return:
        """
        self.enabled = True
        self._app_name = app_name
        self.previous_time = 0  # 上一次性能追踪节点的时间戳，毫秒
        self._link_app_name = link_app_name if link_app_name else []

        if not app_name:
            self._app_name = global_settings().app_name

        if not dispatcher:
            dispatcher = dispatcher_instance()
            console.debug("init application with new dispatcher.")

        self._dispatcher = dispatcher

    @staticmethod
    def singleton_instance(name):
        """one application according to name
        """
        if not name:
            name = global_settings().app_name

        names = str(name).split(';')
        app_name = names[0]
        link_name = names[1:]

        controller = dispatcher_instance()
        instance = Proxy._instances.get(app_name, None)

        if not instance:
            with Proxy._lock:
                instance = Proxy._instances.get(app_name, None)
                if not instance:
                    instance = Proxy(controller, app_name, link_name)
                    Proxy._instances[app_name] = instance

                    console.info("Create new proxy with application name: %s", name)

        return instance

    @property
    def global_settings(self):
        """get the global settings or server settings
        :return:
        """
        return global_settings()

    @property
    def settings(self):
        """
        :return:
        """
        return self._dispatcher.application_settings(self._app_name)

    def activate(self):
        """
        :return:
        """
        self._dispatcher.active_application(self._app_name, self._link_app_name)

    def record_tracker(self, tracker_node):
        """
        """
        self._dispatcher.record_tracker(self._app_name, tracker_node)


def proxy_instance(name=None):
    return Proxy.singleton_instance(name)




"""define the trigger for the specified weapon

"""

import logging
import sys

from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.function_tracker import FunctionTracker, function_trace_wrapper
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance
from tingyun.armoury.trigger.browser_rum import TingYunWSGIBrowserRumMiddleware
from tingyun.armoury.trigger.cross_trace import process_header


console = logging.getLogger(__name__)
TARGET_WEB_APP = "web_app"
TARGET_NOVA_APP = "nova_app"
AGENT_REQUEST_SWITCH = "DISABLE_TING_YUN_AGENT"


class WSGIApplicationResponse(object):

    def __init__(self, tracer, generator):
        self.tracer = tracer
        self.generator = generator

    def __iter__(self):

        try:
            with FunctionTracker(self.tracer, name='Response', group='Python.WSGI'):
                for item in self.generator:
                    yield item
        except GeneratorExit:
            raise
        except:  # Catch all
            self.tracer.record_exception(*sys.exc_info())
            raise

    def close(self):
        """Server/gateway will call this close method to indicate that the task for tracing the request finished.
        :return:
        """

        try:
            if hasattr(self.generator, 'close'):
                self.generator.close()
        except:  # Catch all
            self.tracer.finish_work(*sys.exc_info())
            raise
        else:
            self.tracer.finish_work(None, None, None)


def parse_wsgi_protocol(target, *args, **kwargs):
    """
    :param target: the frame type for wsgi implement
    :param args: the position args for wsgi entrance
    :param kwargs: the key args for wsgi entrance.
    :return:
    """
    if target == TARGET_WEB_APP:
        def wsgi_args(environ, start_response, *args, **kwargs):
            return environ, start_response

        return wsgi_args(*args, **kwargs)
    elif target == TARGET_NOVA_APP:
        def wsgi_args(req, *args, **kwargs):

            return req, args[0]

        return wsgi_args(*args, **kwargs)


def wsgi_wrapper_inline(wrapped, framework="Python", version=None, target=TARGET_WEB_APP):
    """ wrap the uwsgi application entrance
    :param wrapped: the method need to be wrapped
    :param framework: framework of current used web framework
    :param version: version of current used web framework
    :return:
    """
    console.info("wrap the wsgi entrance with framework(%s), version(%s)", framework, version)

    def wrapper(wrapped, instance, args, kwargs):
        """More detail about the argument, read the wrapper doc
        """
        tracker = current_tracker()
        if tracker:
            return wrapped(*args, **kwargs)

        environ, start_response = parse_wsgi_protocol(target, *args, **kwargs)
        disable_agent = environ.get(AGENT_REQUEST_SWITCH, None)
        if disable_agent:
            console.debug("Current trace is disabled with http request environment. %s", environ)
            return wrapped(*args, **kwargs)

        tracker = Tracer(proxy_instance(), environ, framework)
        tracker.start_work()

        # respect the wsgi protocol
        def _start_response(status, response_headers, *args):
            # deal the response header/data
            process_header(tracker, response_headers)
            tracker.deal_response(status, response_headers, *args)
            _write = start_response(status, response_headers, *args)

            return _write

        result = []
        try:
            tracker.set_tracker_name(callable_name(wrapped), priority=1)
            application = function_trace_wrapper(wrapped)
            with FunctionTracker(tracker, name='Application', group='Python.WSGI'):
                result = TingYunWSGIBrowserRumMiddleware(tracker, application, _start_response, environ)()
                # result = application(environ, start_response)
        except:
            tracker.finish_work(*sys.exc_info())
            raise

        return WSGIApplicationResponse(tracker, result)

    return FunctionWrapper(wrapped, wrapper)


def wsgi_application_wrapper(module, object_path, *args):
    """
    :param module:
    :param object_path:
    :return:
    """
    wrap_object(module, object_path, wsgi_wrapper_inline, *args)


def wsgi_application_decorator(framework='xx', version='xx'):
    """
    :param framework: the framework of current use.
    :return:
    """
    framework = 'xx' if framework is None else framework
    version = 'xx' if version is None else version

    def decorator(wrapped):
        """
        :param wrapped:
        :return:
        """
        return wsgi_wrapper_inline(wrapped, framework, version)

    return decorator


wsgi_app_wrapper_entrance = wsgi_wrapper_inline



# -*- coding: utf-8 -*-

"""define a detector for detect concurrent module
"""

import logging
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado_4.utils import create_tracker_aware_fxn, obtain_current_tracker

console = logging.getLogger(__name__)


def trace_add_done_callback(wrapped, instance, args, kwargs):
    """
    """
    def _fxn_arg_extractor(fn, *args, **kwargs):
        return fn

    fxn = _fxn_arg_extractor(*args, **kwargs)
    should_trace = not hasattr(fxn, '_previous_object')
    tracker_aware_fxn = create_tracker_aware_fxn(fxn, should_trace=should_trace)

    # If tracker_aware_fxn is None then it is already wrapped, or the fxn is None.
    if tracker_aware_fxn is None:
        return wrapped(*args, **kwargs)

    tracker = obtain_current_tracker()
    tracker_aware_fxn._nb_tracker = tracker

    # We replace the function we call in the callback with the tracker aware version of the function.
    if len(args) > 0:
        args = list(args)
        args[0] = tracker_aware_fxn
    else:
        kwargs['fn'] = tracker_aware_fxn

    return wrapped(*args, **kwargs)


def detect_concurrent(module):
    """
    """
    wrap_function_wrapper(module, 'Future.add_done_callback', trace_add_done_callback)



# -*- coding: utf-8 -*-

import logging
from tingyun.armoury.ammunition.timer import Timer
from tingyun.logistics.warehouse.mq_node import MQNode
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper

console = logging.getLogger(__name__)


class MQTrace(Timer):
    """
    """
    def __init__(self, tracker, vendor, name_type, host, port, byte, name, headers, role, wrapped_info, external_id):
        """
        :param tracker:
        :param vendor: RabbitMQ/AciveMQ
        :param name_type: Exchange/Queue/Topic
        :param host:
        :param port:
        :param byte:
        :param name:Exchange/routing_key
        :param role: Produce/Consume
        """
        super(MQTrace, self).__init__(tracker)
        self.vendor = vendor or "Unknown"
        self.name_type = name_type or "Unknown"
        self.host = host or "Unknown"
        self.port = port or 0
        self.byte = byte or 0
        self.name = name
        self.headers = headers or {}
        self.role = role
        self.wrapped_info = wrapped_info
        self.external_id = external_id

        # 防止mq驱动自动生成的queue name
        if str(name).startswith('amq.'):
            self.name = 'Temp'
        elif '.' in name:
            self.name_type = 'Topic'

    def create_node(self):
        """
        :return:
        """
        return MQNode(vendor=self.vendor, name_type=self.name_type, host=self.host, port=self.port, byte=self.byte,
                      name=self.name, headers=self.headers, role=self.role, wrapped_info=self.wrapped_info,
                      duration=self.duration, exclusive=self.exclusive, external_id=self.external_id,
                      start_time=self.start_time, end_time=self.end_time, children=self.children)

    def terminal_node(self):
        return False



# -*- coding: utf-8 -*-

import inspect

from tingyun.packages.wrapt.wrappers import wrap_object
from tingyun.packages.wrapt.wrappers import (FunctionWrapper as _FunctionWrapper, ObjectProxy as _ObjectProxy,
                                             BoundFunctionWrapper as _BoundFunctionWrapper)


class TingYunObjectWrapperBase(_ObjectProxy):
    """使用细则参考 http://wrapt.readthedocs.io/en/latest/
    """
    @property
    def _previous_object(self):
        """当一个对象wrapped的时候，可通过该方式判断是否需要再次wrap,同时，如果需要，也可将被wrap的对象返回
        :return:
        """
        try:
            return self._self_previous_object
        except AttributeError:
            self._self_previous_object = getattr(self.__wrapped__, '_previous_object', self.__wrapped__)
            return self._self_previous_object


# _TYBoundFunctionWrapper/FunctionWrapper/function_wrapper定义以及使用，参考wrapt原生的使用文档
# ########################################################################################
class TyBoundFunctionWrapper(TingYunObjectWrapperBase, _BoundFunctionWrapper):
    pass


class TyFunctionWrapper(TingYunObjectWrapperBase, _FunctionWrapper):
    __bound_function_wrapper__ = TyBoundFunctionWrapper


def function_wrapper(wrapper):
    def _wrapper(wrapped, instance, args, kwargs):
        target_wrapped = args[0]
        if instance is None:
            target_wrapper = wrapper
        elif inspect.isclass(instance):
            target_wrapper = wrapper.__get__(None, instance)
        else:
            target_wrapper = wrapper.__get__(instance, type(instance))

        return TyFunctionWrapper(target_wrapped, target_wrapper)

    return TyFunctionWrapper(wrapper, _wrapper)


def wrap_function_wrapper(module, name, wrapper):
    return wrap_object(module, name, TyFunctionWrapper, (wrapper,))
# ########################################################################################



# -*- coding: utf-8 -*-

"""define redis weapon for detect the redis operation information.

"""

from tingyun.armoury.ammunition.function_tracker import wrap_function_trace
from tingyun.armoury.ammunition.redis_tracker import wrap_redis_trace

rewrite_command = ['setex', 'lrem', 'zadd']
basic_command = [
    'append', 'bgrewriteaof', 'bgsave', 'bitcount', 'bitop', 'bitpos', 'blpop', 'brpop', 'brpoplpush', 'client_getname',
    'client_kill', 'client_list', 'client_setname', 'config_get', 'config_resetstat', 'config_rewrite', 'config_set',
    'connection_pool', 'dbsize', 'debug_object', 'decr', 'delete', 'dump', 'echo', 'eval', 'evalsha', 'execute_command',
    'exists', 'expire', 'expireat', 'flushall', 'flushdb', 'from_url', 'get', 'getbit', 'getrange', 'getset', 'hdel',
    'hexists', 'hget', 'hgetall', 'hincrby', 'hincrbyfloat', 'hkeys', 'hlen', 'hmget', 'hmset', 'hscan', 'hscan_iter',
    'hset', 'hsetnx', 'hvals', 'incr', 'incrby', 'incrbyfloat', 'info', 'keys', 'lastsave', 'lindex', 'linsert', 'llen',
    'lock', 'lpop', 'lpush', 'lpushx', 'lrange', 'lrem', 'lset', 'ltrim', 'mget', 'move', 'mset', 'msetnx',
    'object', 'parse_response', 'persist', 'pexpire', 'pexpireat', 'pfadd', 'pfcount', 'pfmerge', 'ping', 'zunionstore',
    'pipeline', 'psetex', 'pttl', 'publish', 'pubsub', 'randomkey', 'register_script', 'rename', 'zscore',
    'renamenx', 'response_callbacks', 'restore', 'rpop', 'rpoplpush', 'rpush', 'rpushx', 'sadd', 'save',
    'scan', 'scan_iter', 'scard', 'script_exists', 'script_flush', 'script_kill', 'script_load', 'sdiff', 'zrevrank',
    'sdiffstore', 'set', 'set_response_callback', 'setbit', 'setex', 'setnx', 'setrange', 'shutdown', 'zscan',
    'sinter', 'sinterstore', 'sismember', 'slaveof', 'slowlog_get', 'slowlog_len', 'slowlog_reset', 'zscan_iter',
    'smembers', 'smove', 'sort', 'spop', 'srandmember', 'srem', 'sscan', 'sscan_iter', 'strlen', 'substr', 'zrevrange',
    'sunion', 'sunionstore', 'time', 'transaction', 'ttl', 'type', 'unwatch', 'watch', 'zadd', 'zcard',
    'zcount', 'zincrby', 'zinterstore', 'zlexcount', 'zrange', 'zrangebylex', 'zrangebyscore', 'zrank',
    'zrem', 'zremrangebylex', 'zremrangebyrank', 'zremrangebyscore', 'zrevrangebyscore',
    ]


def detect_connection(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, 'Connection') and hasattr(module.Connection, 'connect'):
        wrap_function_trace(module, "Connection.connect", name="connect")


def detect_client_operation(module):
    """
    :param module:
    :return:
    """
    def parse_connect_params(instance, *args, **kwargs):
        """连接参数在client的connection_pool变量中
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        try:
            kw = instance.connection_pool.connection_kwargs
            host, port, db = kw.get('host'), kw.get('port'), kw.get('db')
        except AttributeError:
            host, port, db = "Unknown", "Unknown", 0

        return host, port, db or 0

    if hasattr(module, 'StrictRedis'):
        for command in basic_command:
            if hasattr(module.StrictRedis, command):
                wrap_redis_trace(module, "StrictRedis.%s" % command, command, server=parse_connect_params)

        if hasattr(module, 'Redis'):
            for command in rewrite_command:
                if hasattr(module.Redis, command):
                    wrap_redis_trace(module, "Redis.%s" % command, command, server=parse_connect_params)

    elif hasattr(module, 'Redis'):
        for command in basic_command:
            if hasattr(module.Redis, command):
                wrap_redis_trace(module, "Redis.%s" % command, command, server=parse_connect_params)



import re
import weakref

"""this module implement the database trace tools
"""

_sql_parser_catch = weakref.WeakValueDictionary()


_parse_operation_p = r'(\w+)'
_parse_operation_re = re.compile(_parse_operation_p)
_identifier_re = re.compile('[\',"`\[\]\(\)]*')

_uncomment_sql_p = r'/\*.*?\*/'
_uncomment_sql_re = re.compile(_uncomment_sql_p, re.DOTALL)

_normalize_params_1_p = r'%\([^)]*\)s'
_normalize_params_1_re = re.compile(_normalize_params_1_p)
_normalize_params_2_p = r'%s'
_normalize_params_2_re = re.compile(_normalize_params_2_p)
_normalize_params_3_p = r':\w+'
_normalize_params_3_re = re.compile(_normalize_params_3_p)

_normalize_values_p = r'\([^)]+\)'
_normalize_values_re = re.compile(_normalize_values_p)

_normalize_whitespace_1_p = r'\s+'
_normalize_whitespace_1_re = re.compile(_normalize_whitespace_1_p)
_normalize_whitespace_2_p = r'\s+(?!\w)'
_normalize_whitespace_2_re = re.compile(_normalize_whitespace_2_p)
_normalize_whitespace_3_p = r'(?<!\w)\s+'
_normalize_whitespace_3_re = re.compile(_normalize_whitespace_3_p)

_parse_identifier_1_p = r'"((?:[^"]|"")+)"'
_parse_identifier_2_p = r"'((?:[^']|'')+)'"
_parse_identifier_3_p = r'`((?:[^`]|``)+)`'
_parse_identifier_4_p = r'\[\s*(\S+)\s*\]'
_parse_identifier_5_p = r'\(\s*(\S+)\s*\)'
_parse_identifier_6_p = r'([^\s\(\)\[\],]+)'
# _parse_identifier_6_p = r'([^\s\(\)\[\],]+[^;]+)'

_parse_identifier_p = ''.join(('(', _parse_identifier_1_p, '|', _parse_identifier_2_p, '|', _parse_identifier_3_p, '|',
                               _parse_identifier_4_p, '|', _parse_identifier_5_p, '|', _parse_identifier_6_p, ')'))

_parse_from_p = '\s+FROM\s+' + _parse_identifier_p
_parse_from_re = re.compile(_parse_from_p, re.IGNORECASE)
_parse_into_p = '\s+INTO\s+' + _parse_identifier_p
_parse_into_re = re.compile(_parse_into_p, re.IGNORECASE)
_parse_update_p = '\s*UPDATE\s+' + _parse_identifier_p
_parse_update_re = re.compile(_parse_update_p, re.IGNORECASE)
_parse_table_p = '\s+TABLE\s+' + _parse_identifier_p
_parse_table_re = re.compile(_parse_table_p, re.IGNORECASE)

_parse_call_p = r'\s*CALL\s+(?!\()(\w+)'
_parse_call_re = re.compile(_parse_call_p, re.IGNORECASE)
_parse_show_p = r'\s*SHOW\s+(.*)'
_parse_show_re = re.compile(_parse_show_p, re.IGNORECASE | re.DOTALL)
_parse_set_p = r'\s*SET\s+(.*?)\W+.*'
_parse_set_re = re.compile(_parse_set_p, re.IGNORECASE | re.DOTALL)
_parse_exec_p = r'\s*EXEC\s+(?!\()(\w+)'
_parse_exec_re = re.compile(_parse_exec_p, re.IGNORECASE)
_parse_execute_p = r'\s*EXECUTE\s+(?!\()(\w+)'
_parse_execute_re = re.compile(_parse_execute_p, re.IGNORECASE)
_parse_alter_p = r'\s*ALTER\s+(?!\()(\w+)'
_parse_alter_re = re.compile(_parse_alter_p, re.IGNORECASE)


_int_re = re.compile(r'(?<!:)\b\d+\b')
_single_quotes_p = "'(?:[^']|'')*'"
_double_quotes_p = '"(?:[^"]|"")*"'
_any_quotes_p = _single_quotes_p + '|' + _double_quotes_p

_single_quotes_re = re.compile(_single_quotes_p)
_double_quotes_re = re.compile(_double_quotes_p)
_any_quotes_re = re.compile(_any_quotes_p)
_quotes_default = _single_quotes_re

_quotes_table = {
    'MySQLdb': _any_quotes_re,
    'pymysql': _any_quotes_re,
    'oursql': _any_quotes_re,
}

plugin_db_mapping = {
    "MySQLdb": "MySQL",
    "cx_Oracle": "Oracle",
}


_explain_plan_table = {
    'MySQLdb': ('EXPLAIN', ('select',)),
    'ibm_db_dbi': ('EXPLAIN', ('select', 'insert', 'update', 'delete')),
    'oursql': ('EXPLAIN', ('select',)),
    'pymysql': ('EXPLAIN', ('select',)),
    'postgresql.interface.proboscis.engine2': ('EXPLAIN', ('select', 'insert', 'update', 'delete')),
    'psycopg2': ('EXPLAIN', ('select', 'insert', 'update', 'delete')),
    'psycopg2ct': ('EXPLAIN', ('select', 'insert', 'update', 'delete')),
}


def _extract_identifier(token):
    return _identifier_re.sub('', token).strip().lower()


def _parse_default(sql, regex):
    match = regex.search(sql)
    return match and _extract_identifier(match.group(1)) or ''


def _parse_select(sql, engine):
    m = _parse_from_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_delete(sql, engine):
    m = _parse_from_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_insert(sql, engine):
    m = _parse_into_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_update(sql, engine):
    m = _parse_update_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_create(sql, engine):
    m = _parse_table_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_drop(sql, engine):
    m = _parse_table_re.search(sql)
    return m and next(s for s in m.groups()[1:] if s).lower() or ''


def _parse_call(sql, engine):
    return _parse_default(sql, _parse_call_re)


def _parse_show(sql, engine):
    return _parse_default(sql, _parse_show_re)


def _parse_set(sql, engine):
    return _parse_default(sql, _parse_set_re)


def _parse_exec(sql, engine):
    return _parse_default(sql, _parse_exec_re)


def _parse_execute(sql, engine):
    return _parse_default(sql, _parse_execute_re)


def _parse_alter(sql, engine):
    return _parse_default(sql, _parse_alter_re)


_operation_table = {
    'select': _parse_select,
    'delete': _parse_delete,
    'insert': _parse_insert,
    'update': _parse_update,
    'create': _parse_create,
    'drop': _parse_drop,
    'call': _parse_call,
    'show': _parse_show,
    'set': _parse_set,
    'exec': _parse_exec,
    'execute': _parse_execute,
    'alter': _parse_alter,
}


def _parse_operation(sql, engine):
    match = _parse_operation_re.search(sql)
    operation = match and match.group(1).lower() or ''
    return operation if operation in _operation_table else ''


def _parse_table(sql, engine, operation):
    parse = _operation_table.get(operation, None)
    return parse and parse(sql, engine) or ''


def _uncomment_sql(sql, engine):
    return _uncomment_sql_re.sub('', sql)


def _engine_name(engine):
    return engine and hasattr(engine, '__name__') and engine.__name__ or engine


def _obfuscate_sql(sql, engine=None):
    """
    :param sql:
    :param engine:
    :return:
    """
    name = _engine_name(engine)
    quotes_re = _quotes_table.get(name, _quotes_default)

    # Substitute quoted strings first.
    sql = quotes_re.sub('?', sql)

    # Replace straight integer values. This will pick up
    # integers by themselves but also as part of floating point
    # numbers. Because of word boundary checks in pattern will
    # not match numbers within identifier names.
    sql = _int_re.sub('?', sql)

    return sql


def _normalize_sql(sql, engine):
    """
    :param sql:
    :param engine:
    :return:
    """
    sql = _normalize_params_1_re.sub('?', sql)

    # Collapse any parenthesised set of values to a single value.
    sql = _normalize_values_re.sub('(?)', sql)

    # Convert '%s', ':1' and ':name' param styles to '?'.
    sql = _normalize_params_2_re.sub('?', sql)
    sql = _normalize_params_3_re.sub('?', sql)

    # Strip leading and trailing white space.
    sql = sql.strip()

    # Collapse multiple white space to single white space.
    sql = _normalize_whitespace_1_re.sub(' ', sql)

    # Drop spaces adjacent to identifier except for case where
    # identifiers follow each other.
    sql = _normalize_whitespace_2_re.sub('', sql)
    sql = _normalize_whitespace_3_re.sub('', sql)

    return sql


def _explain_plan(sql, engine, connect_params, cursor_params, execute_params):
    """
    :param sql:
    :param engine:
    :param connect_params:
    :param cursor_params:
    :param execute_params:
    :return:
    """
    if engine is None:
        return None

    if type(engine) == type(''):
        return None

    name = _engine_name(engine)

    if name is None:
        return None

    if connect_params is None or cursor_params is None or execute_params is None:
        return None
    if cursor_params is None:
        return None
    if execute_params is None:
        return None

    command, operations = _explain_plan_table.get(name, (None, None))
    if not command:
        return None

    operation = _parse_operation(sql, engine)
    if operation not in operations:
        return None

    query = '%s %s' % (command, sql)
    args, kwargs = connect_params
    try:
        connection = engine.connect(*args, **kwargs)
        try:
            args, kwargs = cursor_params
            cursor = connection.cursor(*args, **kwargs)
            try:
                args, kwargs = execute_params
                cursor.execute(query, *args, **kwargs)
                columns = []
                if cursor.description:
                    for column in cursor.description:
                        columns.append(column[0])
                rows = cursor.fetchall()
                if not columns and not rows:
                    return None

                db_name = plugin_db_mapping.get(name, "Unknown")
                return {"dialect": db_name, "keys": columns, "values": [[str(t) for t in r] for r in rows]}
            except Exception:
                pass
            finally:
                cursor.close()
        finally:
            try:
                connection.rollback()
            except (AttributeError, engine.NotSupportedError):
                pass
            connection.close()
    except Exception:
        pass

    return None


class SQLParser(object):
    def __init__(self, sql, engine=None):
        self.sql = sql
        self.engine = engine

        self._operation = None
        self._table = None
        self._uncommented = None
        self._obfuscated = None
        self._normalized = None
        self._identifier = None

    @property
    def operation(self):
        if self._operation is None:
            self._operation = _parse_operation(self.uncommented, self.engine)

        return self._operation

    @property
    def table(self):
        if self._table is None:
            self._table = _parse_table(self.uncommented, self.engine, self.operation)

        return str(self._table).split(';')[0]

    @property
    def uncommented(self):
        if self._uncommented is None:
            self._uncommented = _uncomment_sql(self.sql, self.engine)

        return self._uncommented

    @property
    def obfuscated(self):
        if self._obfuscated is None:
            self._obfuscated = _obfuscate_sql(self.uncommented, self.engine)

        return self._obfuscated

    @property
    def normalized(self):
        if self._normalized is None:
            self._normalized = _normalize_sql(self.obfuscated, self.engine)

        return self._normalized

    @property
    def identifier(self):
        if self._identifier is None:
            self._identifier = hash(self.normalized)
        return self._identifier

    def formatted(self, sql_format):
        if sql_format == 'off':
            return ''

        elif sql_format == 'raw':
            return self.sql

        else:
            return self.obfuscated

    def explain_plan(self, connect_params, cursor_params, execute_params):
        """do sql explain.
        :param connect_params:
        :param cursor_params:
        :param execute_params:
        :return:
        """
        return _explain_plan(self.sql, self.engine, connect_params, cursor_params, execute_params)


def sql_parser(sql, engine):
    key = (sql, engine.__name__)
    result = _sql_parser_catch.get(key, None)

    if result is not None:
        return result

    result = SQLParser(sql, engine)
    _sql_parser_catch[key] = result

    return result


"""this module is implement the data input read for wsgi

"""
import sys
import threading
import logging

from tingyun.armoury.ammunition.function_tracker import FunctionTracker, wrap_function_trace
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.tracker import current_tracker


middleware_detected_lock = threading.Lock()
console = logging.getLogger(__name__)


def should_ignore(exc, value, tb, ignore_status_code=[]):
    from django.http import Http404

    if isinstance(value, Http404) and 404 in ignore_status_code:
        return True

    return False


def _do_wrap_middleware(middleware, detect_name=False):
    """
    :param middleware: the middleware list in django setting
    :return:
    """
    def wrapper(wrapped):
        name = callable_name(wrapped)

        def wrapper(wrapped, instance, args, kwargs):
            tracker = current_tracker()
            if tracker is None:
                return wrapped(*args, **kwargs)

            before_name = "%s.%s" % (tracker.name, tracker.group)
            with FunctionTracker(tracker, name=name):
                try:
                    return wrapped(*args, **kwargs)
                finally:
                    after_name = "%s.%s" % (tracker.name, tracker.group)
                    if before_name == after_name and detect_name:
                        tracker.set_tracker_name(name, priority=2)

        return FunctionWrapper(wrapped, wrapper)

    for wrapped in middleware:
        yield wrapper(wrapped)


def middleware_wrapper_intermediary(handler, *args, **kwargs):
    """Avoiding two threads deal it in same time
    :param handler: the base handler of http
    :return:
    """
    global middleware_detected_lock
    if not middleware_detected_lock:
        return

    lock = middleware_detected_lock
    lock.acquire()
    if not middleware_detected_lock:
        lock.release()
        return

    middleware_detected_lock = None
    try:
        # for inserting RUM header and footer. indicate it's wrapped and timed as well
        if hasattr(handler, '_response_middleware'):
            pass

        if hasattr(handler, '_request_middleware'):
            handler._request_middleware = list(_do_wrap_middleware(handler._request_middleware, True))

        if hasattr(handler, '_view_middleware'):
            handler._view_middleware = list(_do_wrap_middleware(handler._view_middleware, True))

        if hasattr(handler, '_template_response_middleware'):
            handler._template_response_middleware = list(_do_wrap_middleware(handler._template_response_middleware))

        if hasattr(handler, '_response_middleware'):
            handler._response_middleware = list(_do_wrap_middleware(handler._response_middleware))

        if hasattr(handler, '_exception_middleware'):
            handler._exception_middleware = list(_do_wrap_middleware(handler._exception_middleware))
    finally:
        lock.release()


def middleware_wrapper(wrapped, wrapper_function):
    """
    :return:
    """
    def dynamic_wrapper(wrapped, instance, args, kwargs):
        """
        """
        result = wrapped(*args, **kwargs)

        if instance is not None:
            wrapper_function(instance, *args, **kwargs)
        else:
            wrapper_function(*args, **kwargs)
        return result

    return FunctionWrapper(wrapped, dynamic_wrapper)


def trace_middleware_wrapper(module, object_path):
    """embed the middleware tracker to the django middleware
    :param module:
    :param object_path:
    :return:
    """
    return wrap_object(module, object_path, middleware_wrapper, (middleware_wrapper_intermediary,))


def trace_view_dispatch(module, object_path):
    """django view dispatch trace entrance.
    :param module:
    :param object_path:
    :return:
    """
    def view_dispatch_name(instance, request, *args, **kwargs):
        """The following params in django.views.generic.base.View.dispatch
        :param instance:
        :param request:
        :param args:
        :param kwargs:
        :return:
        """
        if request.method.lower() in instance.http_method_names:
            handler = getattr(instance, request.method.lower(), instance.http_method_not_allowed)
        else:
            handler = instance.http_method_not_allowed

        name = callable_name(handler)
        return name

    return wrap_function_trace(module, object_path, name=view_dispatch_name)


def trace_template_block_render(module, object_path):
    """ wrap the django template load tags times.
    :param module:
    :param object_path:
    :return:
    """
    def block_name(instance, *args, **kwargs):
        """ The following params in django.template.loader_tags.BlockNode.render
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        return instance.name or "default-template"

    return wrap_function_trace(module, object_path, name=block_name, group='Template.Block')


def trace_django_template(module, object_path):
    """
    :param modules:
    :param object_path:
    :return:
    """

    def template_name(instance, *args, **kwargs):
        """ The following params in django.template.base.Template._render/Template.render
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        return instance.name or "default-template"

    return wrap_function_trace(module, object_path, name=template_name, group='Template/Render')


def wrap_view_handler(wrapped, priority=3):
    """
    :param wrapped:
    :param priority:
    :return:
    """
    # views handler maybe called twice on same ResolverMatch. so mark it
    if hasattr(wrapped, '_self_django_view_handler_wrapped'):
        return wrapped

    name = callable_name(wrapped)

    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        tracker.set_tracker_name(name, priority=priority)
        with FunctionTracker(tracker, name=name):
            try:
                return wrapped(*args, **kwargs)
            except:  # Catch all
                tracker.record_exception(ignore_errors=should_ignore)
                raise

    result = FunctionWrapper(wrapped, wrapper)
    result._self_django_view_handler_wrapped = True

    return result


def trace_django_urlresolvers(wrapped):
    """
    :param wrapped:
    :return:
    """
    name = callable_name(wrapped)

    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        if hasattr(tracker, '_self_django_url_resolver_wrapped'):
            return wrapped(*args, **kwargs)

        # mark the top level(views maybe has inline local url resolver operate) url resolver. and use it as the
        tracker._self_django_url_resolver_wrapped = True

        def _wrapped(path):
            with FunctionTracker(tracker, name=name, label=path):
                result = wrapped(path)

                if isinstance(type(result), tuple):
                    callback, callback_args, callback_kwargs = result
                    result = (wrap_view_handler(callback, priority=4), callback_args, callback_kwargs)
                else:
                    result.func = wrap_view_handler(result.func, priority=4)

                return result

        try:
            return _wrapped(*args, **kwargs)
        finally:
            del tracker._self_django_url_resolver_wrapped

    return FunctionWrapper(wrapped, wrapper)


def trace_urlresolvers_resolve_xxx(wrapped, priority):
    """
    :param wrapped:
    :param priority:
    :return:
    """
    name = callable_name(wrapped)

    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        with FunctionTracker(tracker, name=name):
            callback, param_dict = wrapped(*args, **kwargs)
            return wrap_view_handler(callback, priority=priority), param_dict

    return FunctionWrapper(wrapped, wrapper)


def uncaught_exception_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, instance, args, kwargs):
        name = callable_name(wrapped)
        tracker = current_tracker()

        if tracker is None:
            return wrapped(*args, **kwargs)

        def _wrapped(request, resolver, exc_info):
            tracker.set_tracker_name(name, priority=1)
            tracker.record_exception(*exc_info)

            try:
                return wrapped(request, resolver, exc_info)
            except:  # Catch all
                tracker.record_exception(*sys.exc_info())
                raise

        with FunctionTracker(tracker, name=name):
            return _wrapped(*args, **kwargs)

    return FunctionWrapper(wrapped, wrapper)


def trace_uncaught_exception(module, object_path):
    """
    :param module:
    :param object_path:
    :return:
    """
    return wrap_object(module, object_path, uncaught_exception_wrapper)



# -*- coding: utf-8 -*-

import time
import logging
from tingyun.armoury.ammunition.tornado_4.utils import obtain_current_tracker
from tingyun.armoury.ammunition.tornado.wrappers import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado.util import is_websocket, generate_tracer
from tingyun.armoury.ammunition.external_tracker import process_cross_trace

console = logging.getLogger(__name__)


def trace_http_server_headers_received(wrapped, instance, args, kwargs):
    """ instance为HTTPServerRequest实例, 大部分应用都是实现web.Application实例，此时如果有读到数据，则既可会初始化Handler，
    所以从此刻开始算请求开始比，在httpserver._ServerRequestAdapter.headers_received开始算较精确

    另外一种是，用户使用adapter模式时，会从httpserver._ServerRequestAdapter.headers_received开始初始化Handler
    :return:
    """
    # 收到头信息后即刻初始化HTTPServerRequest，所以需要先调用原始函数才能对探针进行操作
    result = wrapped(*args, **kwargs)
    tracker = None if is_websocket(instance) else generate_tracer(instance)

    if tracker:
        tracker.start_time = getattr(instance, "_start_time", int(time.time()))

    return result


def trace_http_headers_init_(wrapped, instance, args, kwargs):
    """
    :return:
    """
    result = wrapped(*args, **kwargs)
    tracker = obtain_current_tracker()
    if tracker and hasattr(instance, 'X-Tingyun-Tx-Data'):
        tracker._called_traced_data = eval(instance.headers.get("X-Tingyun-Tx-Data", '{}'))

    process_cross_trace(instance)
    return result


def detect_http_util(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "HTTPServerRequest.__init__", trace_http_server_headers_received)
    wrap_function_wrapper(module, "HTTPHeaders.__init__", trace_http_headers_init_)



import os
import sys
import imp

from tingyun.config.start_log import log_bootstrap
from tingyun.logistics.mapper import ENV_CONFIG_FILE


log_bootstrap('Ting Yun Bootstrap = %s' % __file__)
log_bootstrap('working_directory = %r' % os.getcwd())
 
# manual load the customize module, just avoid the python module system return something cached when pythonpath changed
boot_directory = os.path.dirname(__file__)
root_directory = os.path.dirname(boot_directory)
log_bootstrap("root dir is: %s" % boot_directory)

path = list(sys.path)
if boot_directory in path:
    del path[path.index(boot_directory)]

# load the other sitecustomize module, which defined in other packages.
try:
    (filename, pathname, description) = imp.find_module('sitecustomize', path)
except ImportError:
    pass
else:
    imp.load_module('sitecustomize', filename, pathname, description)

config_file = os.environ.get(ENV_CONFIG_FILE, None)
log_bootstrap("get config  %s" % config_file, close=True)

if config_file is None:
    private_config = '/opt/tingyun.ini'
    if os.path.isfile(private_config):
        config_file = private_config
        log_bootstrap('used for specified config file[%s] in emergency!' % config_file, close=True)

if config_file is not None:
    if root_directory not in sys.path:
        sys.path.insert(0, root_directory)

    import tingyun.startup
    # Finally startup the world war.
    tingyun.startup.preheat_fight(config_file=config_file)



from __future__ import print_function
import os

from tingyun.config.settings import global_settings
from tingyun.logistics.mapper import ENV_CONFIG_FILE, CONFIG_ITEM
from tingyun.logistics.exceptions import CommandlineParametersException

try:
    import ConfigParser  # for python2.x
except ImportError:
    import configparser as ConfigParser  # for python3.x

_config_parser = ConfigParser.RawConfigParser()


class Command(object):
    """
    """
    def __init__(self):
        """
        """
        self.name = "check-config"  # order format with commandline
        self.options = '[filename]'
        self.description = "Default point to environment %s" % ENV_CONFIG_FILE

    def internal_tips(self, msgs, level="Errors"):
        """
        :return:
        """
        print("----------------------%s------------------------" % level)
        for msg in msgs:
            print("  %s\n" % msg)

        print("----------------------End---------------------------\n")

    def execute(self, args):
        """
        :param args:
        :return:
        """
        errors = []
        warnings = []
        config_file = os.environ.get(ENV_CONFIG_FILE, "")

        if (len(args) > 0) and len(args) != 1:
            raise CommandlineParametersException()

        if 1 == len(args):
            config_file = args[0]

        env_msg = ''
        if 0 == len(args):
            env_msg = "please set the environment variable(%s) for agent config file." % ENV_CONFIG_FILE

        if not os.path.exists(config_file):
            errors.append("Config file is not specified. Or " + env_msg)
            self.internal_tips(errors)
            raise CommandlineParametersException()

        if not _config_parser.read(config_file):
            errors.append("Errors: unable to read the config file. please check the file permission.")
            self.internal_tips(errors)
            raise CommandlineParametersException()

        # check the config detail.
        print("Use config file: %s" % config_file)
        default_settings = global_settings()
        multi_attr = {"action_tracer.log_sql": default_settings.action_tracer.log_sql}
        for item in CONFIG_ITEM:
            ret = self._process_setting(item["section"], item["key"], 'get', item["mapper"])
            if ret[0] < 0:
                errors.append(ret[2])
                break

            if 0 == ret[0] and 'log_file' != item["key"] and 'license_key' != item["key"]:
                if item["key"] in multi_attr:
                    v = multi_attr[item["key"]]
                else:
                    v = getattr(default_settings, item["key"]) if 'log_level' != item["key"] else "logging.DEBUG"
                warnings.append(ret[2] + " Use default value [%s] instead." % v)
                continue

            if 'log_file' == item["key"] and not ret[1]:
                warnings.append("config option <log_file> is not defined, agent log will output to stderr.")
                continue

            if 'log_file' == item["key"] and ret[1]:
                try:
                    with open(ret[1], "a+") as fd:
                        fd.write("agent check log file config input message.\n")
                except Exception as _:
                    warnings.append("Current user[%s] not allowed to access the log file[%s]."
                                    % (os.environ["USER"], ret[1]))

            if 'license_key' == item["key"] and not ret[1]:
                errors.append("config option <license_key> is not defined, agent will not work well.")
                continue

        if warnings:
            self.internal_tips(warnings, "Warning")

        if errors:
            self.internal_tips(errors)
            raise CommandlineParametersException()

        if not errors:
            print("\nValidate agent config file success!!")

    def _process_setting(self, section, option, getter='get', mapper=None):
        """
        Return: [errorCode, value, errMsg]

        """
        try:
            map_value = ""
            value = getattr(_config_parser, getter)(section, option)
            if mapper:
                    map_value = mapper(value)

            # use the default value in settings
            if map_value is None:
                return [0, None, "%s=%s, %s is not supported." % (option, value, value)]

            return [1, value, ""]

        except ConfigParser.NoSectionError:
            return [-1, None, "Section [%s] is not specified." % section]
        except ConfigParser.NoOptionError:
            return [0, None, "Option <%s> is not exist." % option]
        except Exception as err:
            return [-1, None, err]



# -*- coding: utf-8 -*-
from __future__ import print_function

with_setuptools = False

try:
    from setuptools import setup
    with_setuptools = True
except ImportError:
    from distutils.core import setup

from distutils.core import Extension
from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError, DistutilsPlatformError)

import os
import sys

version = __import__('tingyun').get_version()
_copyright = '(C) Copyright 2007-2017 networkbench Inc. All rights reserved.'
packages, package_data = [], {}
root_dir = os.path.dirname(__file__)
tingyun_dir = 'tingyun'
if root_dir != '':
    os.chdir(root_dir)


def split(dir_path, result=None):
    """
    """
    if result is None:
        result = []
    head, tail = os.path.split(dir_path)
    if head == '':
        return [tail] + result
    if head == dir_path:
        return result
    return split(head, [tail] + result)


def process_packages():
    """collect the packages info into package_data
    :return:
    """
    for dir_path, dir_name, file_name in os.walk(tingyun_dir):
        dir_name[:] = [d for d in dir_name if not d.startswith('.') and d != '__pycache__']
        parts = split(dir_path)
        package_name = '.'.join(parts)
        if '__init__.py' in file_name:
            packages.append(package_name)
        elif file_name:
            relative_path = []
            while '.'.join(parts) not in packages:
                relative_path.append(parts.pop())
            relative_path.reverse()
            path = os.path.join(*relative_path)
            package_files = package_data.setdefault('.'.join(parts), [])
            package_files.extend([os.path.join(path, f) for f in file_name])


# package the setup data for build python package
process_packages()
kwargs = dict(
    name="tingyun",
    version=version,
    description="Python application performance monitor client,It's main usage is working with python web application.",
    long_description="Application performance monitor client, which is base on tingyun platform. Agent can monitor "
                     "python web framework/modules performance, such as Django, Tornado, database, external call etc."
                     "In the premise of using less resources, our agent can achieve a variety of module tracking.",
    author="tingyun.com",
    author_email="python@tingyun.com ",
    license=_copyright,
    platforms=['unix', 'linux', 'MacOS'],
    url="http://www.tingyun.com",
    packages=packages,
    package_data={'tingyun': ['tingyun.ini', 'tingyun/tingyun.ini', 'LICENSE', 'packages/requests/cacert.pem']},
    scripts=['scripts/tingyun-admin'],
    classifiers=[
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: Unix',
    ],
)

if with_setuptools:
    kwargs['entry_points'] = {'console_scripts': ['tingyun-admin = tingyun.commander:launch_commanding_elevation']}

if sys.platform == 'win32':
    build_ext_errors = (CCompilerError, DistutilsExecError,  DistutilsPlatformError, IOError)
else:
    build_ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


class BuildExtFailed(Exception):
    pass


class TinYunBuildExtension(build_ext):
    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildExtFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except build_ext_errors:
            raise BuildExtFailed()


def _run_setup(with_extension):
    """
    :param with_extension:
    :return:
    """
    setup_kwargs = dict(kwargs)

    if with_extension:
        setup_kwargs['ext_modules'] = [
            Extension("tingyun.packages.wrapt._wrappers", ["tingyun/packages/wrapt/_wrappers.c"])
        ]

    setup_kwargs['cmdclass'] = dict(build_ext=TinYunBuildExtension)
    setup(**setup_kwargs)


def do_run_setup_install():
    """
    :return:
    """
    with_extensions = True

    # skip the pypy.
    if hasattr(sys, 'pypy_version_info'):
        with_extensions = False
        print('========================================================')

    try:
        _run_setup(with_extensions)
    except Exception as _:
        print(80 * '*')

        print("""
                                  =========================================
                                               WARNING
                                  =========================================
                The optional C extension components of the Python agent could not be compiled.
              Maybe the C compiler is not installed on the system。 The Python agent will be installed without
              the C extensions. ."""
        )

        print("INFO: Trying to build without extensions.")
        print(80 * '*')

        _run_setup(with_extension=False)
        print(80 * '*')
        print("INFO: Only pure Python agent was installed.")

        print(80 * '*')

# Actually run the setup for agent.
do_run_setup_install()




"""This module do some inspect and preparation for armoury and corps

"""
import logging
import os
import sys
import threading
import traceback

try:
    import ConfigParser  # for python2.x
except ImportError:
    import configparser as ConfigParser  # for python3.x

from tingyun.logistics.mapper import ENV_CONFIG_FILE
from tingyun.logistics.exceptions import ConfigurationError
from tingyun.logistics.mapper import map_log_level, map_key_words, map_app_name
from tingyun.config.settings import global_settings
from tingyun.embattle.log_file import initialize_logging
from tingyun.packages.wrapt.importer import register_post_import_hook
from tingyun.embattle.repertoire import defined_repertoire
from tingyun.config.start_log import log_bootstrap

console = logging.getLogger(__name__)


class Embattle(object):
    """
    """
    _lock = threading.Lock()
    _instance = None

    def __init__(self, config_file):
        """
        :param config_file:
        :return:
        """
        self.config_file = config_file
        self.is_embattled = False
        self.valid_embattle = True
        self._config_parser = ConfigParser.RawConfigParser()
        self._settings = global_settings()
        self._inspect_lock = threading.Lock()

        if not config_file:
            self.config_file = os.environ.get(ENV_CONFIG_FILE, None)

        if not self.config_file:
            log_bootstrap('Agent config file is not found, agent start failed.')
            self.valid_embattle = False

        log_bootstrap("get config file %s" % self.config_file)

    @staticmethod
    def singleton_instance(config_file):
        """one application according to name
        """
        instance = Embattle._instance

        if not instance:
            with Embattle._lock:
                instance = Embattle._instance
                if not instance:
                    instance = Embattle(config_file)
                    Embattle._instance = instance

        return instance

    def inspect_prerequisites(self):
        """ Do some prerequisite check for corps operation
        :return:
        """
        log_bootstrap("Inspecting config file %s" % self.config_file, close=True)

        if not self._config_parser.read(self.config_file):
            raise ConfigurationError("Unable to access the config file. %s", self.config_file)

        self._settings.config_file = self.config_file
        self.load_settings()
        self.load_settings('tingyun:exclude', "plugins")
        self.load_settings('tingyun:private', "port")
        self.load_settings('tingyun:private', "host")
        self.load_settings('tingyun:proxy', "proxy_host")
        self.load_settings('tingyun:proxy', "proxy_port")
        self.load_settings('tingyun:proxy', "proxy_user")
        self.load_settings('tingyun:proxy', "proxy_pwd")
        self.load_settings('tingyun:proxy', "proxy_scheme")

        # we can not access the log file and log level from the config parser. it maybe empty.
        initialize_logging(self._settings.log_file, self._settings.log_level)

    def load_settings(self, section='tingyun', option=None, method='get'):
        """ Load the specified settings from the ini config file
        :param section: ini parser section
        :param method: ini parser option operate method
        :return:
        """
        self._settings.config_file = self.config_file

        if option is None:
            self._process_setting(section, 'log_file', method, None)
            self._process_setting(section, 'log_level', method, map_log_level)
            self._process_setting(section, 'license_key', method, None)
            self._process_setting(section, 'enabled', method, map_key_words)
            self._process_setting(section, 'app_name', method, map_app_name)
            self._process_setting(section, 'audit_mode', method, map_key_words)
            self._process_setting(section, 'auto_action_naming', method, map_key_words)
            self._process_setting(section, 'ssl', method, map_key_words)
            self._process_setting(section, 'action_tracer.log_sql', method, map_key_words)
            self._process_setting(section, 'daemon_debug', method, map_key_words, True)
            self._process_setting(section, 'enable_profile', method, map_key_words)
            self._process_setting(section, 'urls_merge', method, map_key_words)
            self._process_setting(section, 'verify_certification', method, map_key_words)

            self._process_setting(section, 'tornado_wsgi_adapter_mode', method, map_key_words)
        else:
            self._process_setting(section, option, method, None, True)

    def _process_setting(self, section, option, method='get', mapper=None, hide=False):
        """ load the settings from config file to the settings objects instance.
        :param section: ini configure file section
        :param option: ini configure file option
        :param method: ini configure file operation method
        :param mapper: the method which mapping key value to the ini config option value
        :param hide: the option value error is can be ignored.
        :return: None
        """
        try:
            value = getattr(self._config_parser, method)(section, option)
            value = value if not mapper else mapper(value)

            # invalid value with mapper func mapping, used default instead
            if value is None:
                return

            target = self._settings
            fields = option.split('.', 1)

            while True:
                if len(fields) == 1:
                    setattr(target, fields[0], value)
                    break
                else:
                    target = getattr(target, fields[0])
                    fields = fields[1].split('.', 1)
        except ConfigParser.NoSectionError:
            if not hide:
                console.debug("No section[%s] in configure file", section)
        except ConfigParser.NoOptionError:
            if not hide:
                console.debug("No option[%s] in configure file", option)
        except Exception as err:
            console.warning("Process config error, section[%s]-option[%s] will use default value instead. %s", err)

    def detector(self, target_module, hook_module, function):
        """
        :param target_module:
        :param hook_module:
        :param function:
        :return:
        """
        def _detect(target_module):
            """
            """
            try:
                getattr(self.importer(hook_module), function)(target_module)
                console.info("Detect hooker %s for target module %s", hook_module, target_module)
            except Exception as _:
                console.warning("error occurred: %s" % traceback.format_exception(*sys.exc_info()))

        return _detect

    def importer(self, name):
        """
        :param name:
        :return:
        """
        __import__(name)
        return sys.modules[name]

    def activate_weapons(self):
        """
        :return:
        """
        exclude = self._settings.plugins
        for name, hooks in defined_repertoire().items():
            if name in exclude:
                console.debug("Ignore the plugin %s", name)
                continue
            for hook in hooks:
                target = hook.get('target', '')
                hook_func = hook.get('hook_func', '')
                hook_module = hook.get('hook_module', '')
                register_post_import_hook(self.detector(target, hook_module, hook_func), target)

    def execute(self):
        """
        :return:
        """
        if not self.valid_embattle:
            return False

        # has been for some prerequisite check for corps operation
        if self.is_embattled:
            console.warning("Agent was initialized before.")
            return False

        with self._inspect_lock:

            if self.is_embattled:
                console.warning('agent was initialized before, skip it now.')
                return False

            log_bootstrap("Stating init agent environment %s." % self._settings)
            self.is_embattled = True

            try:
                self.inspect_prerequisites()
                self.activate_weapons()
            except Exception as err:
                console.error("Errors. when initial the agent system. %s", err)
                return False

            return True


# This must the only entry for holding the corps's embattle
def take_control(config_file=None):
    """
    :param config_file: config file for the corps
    :return:
    """
    return Embattle.singleton_instance(config_file)




"""this module is defined to detect the requests module
"""

import sys
import logging
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import wrap_external_trace, process_cross_trace, parse_parameters
from tingyun.armoury.ammunition.external_tracker import MALFORMED_URL_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CONNECT_EXCEPTION_ERROR_CODE, TIMEOUT_EXCEPTION_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import CLIENT_PROTOCOL_ERROR_CODE, ILLEGAL_RESPONSE_ERROR_CODE
from tingyun.armoury.ammunition.external_tracker import OTHER_ERROR_CODE

# not exist at v2.0.0
try:
    from requests.exceptions import ConnectTimeout
except ImportError:
    class ConnectTimeout(Exception):
        pass

try:
    from requests.exceptions import ReadTimeout
except ImportError:
    class ReadTimeout(Exception):
        pass

try:
    from requests.exceptions import ContentDecodingError
except ImportError:
    class ContentDecodingError(Exception):
        pass

try:
    from requests.exceptions import StreamConsumedError
except ImportError:
    class StreamConsumedError(Exception):
        pass

try:
    from requests.exceptions import InvalidSchema
except ImportError:
    class InvalidSchema(Exception):
        pass

try:
    from requests.exceptions import InvalidURL
except ImportError:
    class InvalidURL(Exception):
        pass

try:
    from requests.exceptions import URLRequired
except ImportError:
    class URLRequired(Exception):
        pass


try:
    from requests.exceptions import ChunkedEncodingError
except ImportError:
    class ChunkedEncodingError(Exception):
        pass

try:
    from requests.exceptions import MissingSchema
except ImportError:
    class MissingSchema(Exception):
        pass

try:
    from requests.exceptions import InvalidSchema
except ImportError:
    class InvalidSchema(Exception):
        pass

try:
    from requests.exceptions import ConnectionError
except ImportError:
    class ConnectionError(Exception):
        pass

try:
    from requests.exceptions import SSLError
except ImportError:
    class SSLError(Exception):
        pass


console = logging.getLogger(__name__)


def wrap_exception(wrapped, action_url, parameters, *args, **kwargs):
    """
    """

    tracker = current_tracker()
    if tracker is None:
        return wrapped(*args, **kwargs)

    url = action_url if not callable(action_url) else action_url(*args, **kwargs)
    params = parameters if not callable(parameters) else parameters(*args, **kwargs)

    http_status = 500
    error_code = 0
    params = parse_parameters(url, params)

    try:
        rtv = wrapped(*args, **kwargs)
    except ReadTimeout:
        error_code = TIMEOUT_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise
    except (ConnectTimeout, ConnectionError, SSLError):
        error_code = CONNECT_EXCEPTION_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise
    except (InvalidURL, URLRequired):
        error_code = MALFORMED_URL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise
    except (InvalidSchema, MissingSchema):
        error_code = CLIENT_PROTOCOL_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise
    except (ChunkedEncodingError, ContentDecodingError, StreamConsumedError):
        error_code = ILLEGAL_RESPONSE_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise
    except Exception:
        error_code = OTHER_ERROR_CODE
        tracker.record_external_error(url, error_code, http_status, sys.exc_info(), params, module_name='requests')
        raise

    if int(getattr(rtv, 'status_code', 0)) != 200:
        tracker.record_external_error(url, error_code, getattr(rtv, 'status_code', 500), sys.exc_info(), params,
                                      module_name='requests')

    return rtv


def detect_requests_sessions(module):
    """
    :param module:
    :return:
    """
    def request_url(instance, method, url, *args, **kwargs):
        """
        """
        return url

    def parse_params(method, url, *args, **kwargs):
        _kwargs = kwargs
        _args = (method, url) + args
        protocol = 'http' if 'https' not in url else 'https'

        if 'headers' in _kwargs:
            ty_headers, external_id = process_cross_trace(_kwargs['headers'], protocol)
            _kwargs['headers'] = ty_headers
        else:
            ty_headers, external_id = process_cross_trace(None, protocol)
            _kwargs['headers'] = ty_headers

        return external_id, _args, _kwargs

    wrap_external_trace(module, 'Session.request', 'requests', request_url, params=parse_params,
                        exception_wrapper=wrap_exception)






"""
"""

import logging

from tingyun.logistics.warehouse.memcache_node import MemcacheNode
from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper
from tingyun.armoury.ammunition.timer import Timer
from tingyun.armoury.ammunition.tracker import current_tracker

console = logging.getLogger(__name__)


class MemcacheTrace(Timer):
    def __init__(self, tracker, host, port, command):
        super(MemcacheTrace, self).__init__(tracker)
        self.host = host
        self.port = port
        self.command = command

    def create_node(self):
        tracker = current_tracker()
        if tracker:
            tracker.memcache_time = self.duration

        return MemcacheNode(command=self.command, children=self.children, start_time=self.start_time, host=self.host,
                            end_time=self.end_time, duration=self.duration, exclusive=self.exclusive, port=self.port)

    def terminal_node(self):
        return True


def memcached_trace_wrapper(wrapped, server, command):
    """
    :return:
    """
    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        if callable(server):
            host, port = server(instance, *args, **kwargs)
        else:
            host, port = "Unknown", 0

        if instance is not None:
            _command = command(instance, *args, **kwargs)
        else:
            _command = command(*args, **kwargs)

        with MemcacheTrace(tracker, host, port, _command):
            return wrapped(*args, **kwargs)

    def literal_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        if tracker is None:
            return wrapped(*args, **kwargs)

        if callable(server):
            host, port = server(instance, *args, **kwargs)
        else:
            host, port = "Unknown", 0

        with MemcacheTrace(tracker, host, port, command):
            return wrapped(*args, **kwargs)

    if callable(command):
        return FunctionWrapper(wrapped, dynamic_wrapper)

    return FunctionWrapper(wrapped, literal_wrapper)


# egg: (memcached, Client.append, get)
def wrap_memcache_trace(module, object_path, command, server=None):
    wrap_object(module, object_path, memcached_trace_wrapper, (server, command))



# -*- coding:utf-8 -*-

"""This module implements data recording and reporting for an application

"""

import json
import logging
import threading
import time
import sys

from tingyun.logistics.transportation.engine import create_connection
from tingyun.logistics.workshop.packager import Packager
from tingyun.armoury.sampler.environment import env_config
from tingyun.armoury.sampler.profile import get_profile_manger
from tingyun.armoury.sampler.sampler import SamplerController
from tingyun.logistics.exceptions import InvalidLicenseException, OutOfDateConfigException, InvalidDataTokenException
from tingyun.logistics.exceptions import RetryDataForRequest, DiscardDataForRequest, ServerIsUnavailable
from tingyun.config.settings import global_settings
from tingyun.packages.wrapt.decorators import synchronized
from tingyun.logistics.mapper import CONSTANCE_OUT_DATE_CONFIG, CONSTANCE_RETRY_DATA, CONSTANCE_SERVER_UNAVAILABLE
from tingyun.logistics.mapper import CONSTANCE_DISCARD_DATA, CONSTANCE_HARVEST_ERROR, CONSTANCE_INVALID_DATA_TOKEN
from tingyun.logistics.mapper import CONSTANCE_LICENSE_INVALID, CONSTANCE_SESSION_NOT_ACTIVED
from tingyun.logistics.mapper import CONSTANCE_INVALID_LICENSE_KEY, MAX_RETRY_UPLOAD_FAILED_TIMES


console = logging.getLogger(__name__)


class Application(object):
    """Real web application property/action/dispatcher holder/controller
    """
    def __init__(self, app_name, linked_app=[]):
        """
        :param app_name: a real python/project application name, usually write in config file
        :param linked_app: linked app name for app
        :return:
        """
        console.info("Init Application with name %s, linked-name %s", app_name, linked_app)

        # if retry upload data failed. will not rollback the data and reset the session
        self.retry_upload_times = 0

        self._app_name = app_name
        self._linked_app = sorted(set(linked_app))
        self._active_session = None
        self.connect_thread = None
        self._is_license_valid = True

        self.sampler = []

        self.connect_retry_time = 60.0
        self._connect_event = threading.Event()
        self._packets_lock = threading.Lock()  # used to lock the core collect data
        self._packager = Packager()

        self._agent_commands_lock = threading.Lock()
        self.profile_manager = get_profile_manger()
        self.profile_status = None

        self.__max_tracker = 2000
        self._tracker_count = 0
        self._last_tracker = 0.0
        self.__data_sampler = []
        self.metric_name_ids = {"actions": {}, "apdex": {}, "components": {}, "general": {}, "errors": {}}

    @property
    def active(self):
        """
        :return:
        """
        return self._active_session is not None

    @property
    def application_config(self):
        """return the configuration of the application, it's downloaded from server witch is merged with uploaded
        settings with server config on server
        :return:
        """
        return self._active_session and self._active_session.config

    def stop_connecting(self):
        """
        :return:
        """
        self._connect_event.set()

    def shutdown_internal_service(self):
        """used to shutdown the internal commands service.
        :return:
        """
        self.profile_manager.shutdown()
        self.stop_sampler()

    def activate_session(self):
        """active a session(background thread) to register the application
        :return:
        """
        # TODO: do some check for agent status
        if self.connect_thread and self.connect_thread.is_alive():
            console.warning("Activate thread is active. maybe something wrong when dispatcher activate application.")
            return False

        self._connect_event.clear()

        self.connect_thread = threading.Thread(target=self.connect_to_headquarters, name="tingYunSessionThread")
        self.connect_thread.setDaemon(True)
        self.connect_thread.start()

        return True

    def connect_to_headquarters(self):
        """Performs the actual registration of the application to server, get server config and set the current app
        settings.
        :return:
        """
        # ensure the main thread get to run first
        time.sleep(0.01)

        while not self._active_session:
            self.retry_upload_times = 0
            settings = global_settings()

            try:
                active_session = create_connection(None, self._app_name, self._linked_app, env_config(), settings)
            except InvalidLicenseException as _:
                console.warning("Invalid license in configuration, agent will stop to work please fix license and"
                                "restart agent again")

                self._is_license_valid = False
                self._connect_event.wait(self.connect_retry_time)
                continue
            except Exception:
                # use the harvest controller signal to control the connection
                console.exception("Connect to agent server failed, Connection will try again in 1 min.")
                self._connect_event.wait(self.connect_retry_time)
                continue
            finally:
                if self._connect_event.isSet():
                    console.info("Agent is shutting down, stop the connection to server now.")
                    return

            if active_session:
                self._is_license_valid = True
                self._connect_event.set()
                self.start_sampler()

                # set the application settings to data engine
                with self._packets_lock:
                    self._active_session = active_session
                    self._packager.reset_packets(self._active_session.config)

    def register_data_sampler(self, sampler, *args):
        """
        :param sampler:
        :param args:
        :return:
        """
        self.__data_sampler.append(SamplerController(sampler, args))

    @synchronized
    def start_sampler(self):
        """
        :return:
        """

        for sampler in self.__data_sampler:
            try:
                console.debug("Starting data sampler in  %s", self._app_name)
                sampler.start()
            except Exception as err:
                console.exception("Exception occurred %s, when start sampler ", err, sampler.name)

    @synchronized
    def stop_sampler(self):
        """
        :return:
        """

        for sampler in self.__data_sampler:
            try:
                sampler.stop()
            except Exception as err:
                console.exception("Exception occurred %s, when stop sampler ", err, sampler.name)

    def harvest(self, last_harvest, current_harvest, shutdown=False):
        """Performs a harvest, reporting aggregated data for the current reporting period to the server.
        :return:
        """
        ret_code = 0

        # controller should ignore session/license error code, because the connecting thread doing/did it
        if not self._active_session:
            console.warning("Application not registered to server yet, skip harvest data.")
            return CONSTANCE_SESSION_NOT_ACTIVED, self._app_name, self._linked_app

        if not self._is_license_valid:
            console.debug("The license is invalid, skip harvest data.")
            return CONSTANCE_LICENSE_INVALID, self._app_name, self._linked_app

        with self._packets_lock:
            self._tracker_count = 0
            self._last_tracker = 0.0
            stat = self._packager.packets_snapshot()

        config = self.application_config

        for sampler in self.__data_sampler:
            if not config.enabled:
                break

            try:
                for s in sampler.metrics(current_harvest - last_harvest):
                    stat.record_time_metrics(s)
            except Exception as err:
                console.exception("Sample the data occurred errors. %s", err)

        try:
            # send metric data even if disabled by server
            performance_metric = self.get_performance_metric(stat, last_harvest, current_harvest, config.audit_mode)
            if config.audit_mode:
                console.info("Agent capture the performance data %s", performance_metric)

            result = self._active_session.send_performance_metric(performance_metric)

            # is application is disabled by server. send empty data and ignore all of other data. even if it is not
            # captured.
            if not config.enabled:
                stat.reset_metric_packets()
                return

            self.process_metric_id(result, config.daemon_debug)

            if config.error_collector.enabled:
                error_trace = self.get_error_trace_data(stat)
                external_error_trace = self.get_external_error_trace(stat)

                if config.audit_mode:
                    console.info("Agent capture the error trace data %s", error_trace)
                    console.info("Agent capture the external error trace data %s", external_error_trace)

                self._active_session.send_error_trace(error_trace)
                self._active_session.send_external_error_trace(external_error_trace)

            if config.action_tracer.enabled:
                slow_action_data = stat.action_trace_data()

                if config.audit_mode:
                    console.info("Agent capture the slow action data %s", slow_action_data)

                self._active_session.send_action_trace(slow_action_data)

            if config.action_tracer.slow_sql:
                slow_sql_data = stat.slow_sql_data()
                if config.audit_mode:
                    console.info("Agent capture the slow sql data %s", slow_sql_data)

                self._active_session.send_sql_trace(slow_sql_data)

            stat.reset_metric_packets()

            # get the commands and execute it.
            self.process_headquarters_command()
            self.send_profile_data(config.audit_mode)

            if shutdown:
                self.shutdown_internal_service()
        except OutOfDateConfigException as _:
            # need to reset the connection
            self._active_session = None
            ret_code = CONSTANCE_OUT_DATE_CONFIG
            console.info("Config changed in server, reset the connect now.")
        except InvalidDataTokenException as _:
            # if license key when upload the data ,then reset the session
            self._active_session = None
            ret_code = CONSTANCE_INVALID_DATA_TOKEN
            console.info("Data token is valid, register the application %s again now", self._app_name)
        except InvalidLicenseException as _:
            # if invalid license error occurred when upload data. current data will be lost. And dispatcher will reset
            # session.
            self._active_session = None
            ret_code = CONSTANCE_INVALID_LICENSE_KEY
        except RetryDataForRequest as _:
            console.warning("This exception indicates server service can not touched. if this error continues. please "
                            "report to us for further investigation. thank u")

            # if upload try times more than max, reset the session and raise invalid data token, thant will trigger
            # reconnection to data collector. This will avoid some issue for memory use because of data rollback.
            ret_code = CONSTANCE_RETRY_DATA
            if self.retry_upload_times >= MAX_RETRY_UPLOAD_FAILED_TIMES:
                self._active_session = None
                ret_code = CONSTANCE_INVALID_DATA_TOKEN

                console.warning("Upload data to the dc failed with max retry %s", self.retry_upload_times)
            else:
                self.retry_upload_times += 1
                with self._packets_lock:
                    try:
                        self._packager.rollback(stat)
                    except Exception as err:
                        console.warning("rollback(%s) performance data failed. %s", self.retry_upload_times, err)

        except Exception as err:
            ret_code = CONSTANCE_HARVEST_ERROR
            console.exception("Some error occurred in agent code. if this error continues. "
                              "please report to us for further investigation. thank u.")
            console.exception("%s", err)

        return ret_code, self._app_name, self._linked_app
        
    def record_tracker(self, tracker):
        """
        """
        if not self._active_session or not self._packager.settings:
            console.debug("Agent server is disconnected, tracker data will be dropped.")
            return False

        try:
            stat = self._packager.create_data_zone()
            stat.record_tracker(tracker)

            self._tracker_count += 1
            self._last_tracker = tracker.end_time

            with self._packets_lock:
                self._packager.merge_metric_packets(stat)
        except Exception as err:
            console.exception("Unexpected error occurred when record tracker data, that's maybe the agent code issue "
                              "if this continues, please report to us for further investigation. %s", err)

        return False

    def process_metric_id(self, metric_ids, debug_mode=False):
        """keep the metric id in the memory for replace the key
        :param metric_ids:the metric ids download from server.
        :return:
        """
        if not metric_ids or debug_mode:
            return self.metric_name_ids

        if "actions" in metric_ids:
            for item in metric_ids["actions"]:
                key = item[0]["name"].encode("utf8")
                self.metric_name_ids["actions"][key] = item[1]

        if "apdex" in metric_ids:
            for item in metric_ids["apdex"]:
                key = item[0]["name"].encode("utf8")
                self.metric_name_ids["apdex"][key] = item[1]

        if "general" in metric_ids:
            for item in metric_ids["general"]:
                key = item[0]["name"].encode("utf8")
                self.metric_name_ids["general"][key] = item[1]

        if "errors" in metric_ids:
            for item in metric_ids["errors"]:
                key = item[0]["name"].encode("utf8")
                self.metric_name_ids["errors"][key] = item[1]

        if "components" in metric_ids:
            for item in metric_ids["components"]:
                key = "%s:%s" % (item[0]["name"], item[0]["parent"])
                key = key.encode("utf8")
                self.metric_name_ids["components"][key] = item[1]

        return self.metric_name_ids

    def get_performance_metric(self, stat, last_harvest, current_harvest, audit_mode=False):
        """
        :param stat:
        :return:
        """
        # disable the id mechanism
        metric_name_ids = self.metric_name_ids
        if audit_mode:
            metric_name_ids = {"actions": {}, "apdex": {}, "components": {}, "general": {}, "errors": {}}

        # if application is disabled. should send some heartbit data.
        if not self.application_config.enabled:
            performance = {
                "type": "perfMetrics",
                "timeFrom": int(last_harvest),
                "timeTo": int(current_harvest),
                "interval": int(current_harvest - last_harvest),
                "actions": [],
                "apdex": [],
                "components": [],
                "general": [],
                "errors": stat.error_packets(metric_name_ids["errors"])
            }

            return performance

        performance = {
            "type": "perfMetrics",
            "timeFrom": int(last_harvest),
            "timeTo": int(current_harvest),
            "interval": int(current_harvest - last_harvest),
            "actions": stat.action_metrics(metric_name_ids["actions"]),
            "apdex": stat.apdex_data(metric_name_ids["apdex"]),
            "components": stat.component_metrics(metric_name_ids["components"]),
            "general": stat.general_trace_metric(metric_name_ids["general"]),
            "errors": stat.error_packets(metric_name_ids["errors"]),
        }

        if 0 != len(self._active_session.config.quantile):
            performance["config"] = {"nbs.quantile": json.dumps(self._active_session.config.quantile_org)}

        return performance

    def get_error_trace_data(self, stat):
        """
        :return:
        """
        error_trace = {
            "type": "errorTraceData",
            "errors": stat.error_trace_data()
        }

        # no error data recorded return None as mark.
        if 0 == len(error_trace['errors']):
            error_trace = []

        return error_trace

    def get_external_error_trace(self, stat):
        """
        :param stat:
        :return:
        """
        error_data = stat.external_error_data()
        error_trace = {
            "type": "externalErrorTraceData",
            "errors": error_data
        }

        # no error data recorded return None as mark.
        if 0 == len(error_data):
            error_trace = []

        return error_trace

    def process_headquarters_command(self):
        """get the command from agent server, and start the command.
        :return:
        """
        # use the lock for only sure one processes on the agent command

        with self._agent_commands_lock:
            for cmd in self._active_session.request_agent_commands():
                console.info("Processing command %s", cmd)

                cmd_id = None
                if 'StopProfiler' not in cmd['name']:
                    cmd_id = cmd['id']

                cmd_name = cmd['name']
                cmd_args = cmd['args']

                cmd_handler = getattr(self, "cmd_%s" % cmd_name, None)
                if cmd_handler is None:
                    console.info("Agent dose not support command %s", cmd_name)
                    continue

                cmd_handler(cmd_id, cmd_args)

    def cmd_StartProfiler(self, cid, args):
        """start profile function, we define this just for adapt to auto match the commands
        :param cid: command id
        :param args: command args
        :return: None
        """
        profile_id = args["profileId"]
        duration = args['duration']
        interval = args['interval']

        if self.profile_status:
            return

        if not hasattr(sys, "_current_frames"):
            """
            """
            console.warning("The current Python interpreter being used is not support thread profiling. For more help "
                            "about the thread profiling, please contact our support.")
            return

        self.profile_status = self.profile_manager.start_profile(cid, self._app_name, profile_id, duration, interval)
        if not self.profile_status:
            console.warning("profiler is running now, so skip current profile command. %s", args)
            return

        console.info("Starting thread profile for application %s, with duration %ss,  interval %sms",
                     self._app_name, duration, interval)

    def cmd_StopProfiler(self, cid, args):
        """ stop the profile commands
        """
        profile_id = args['profileId']
        profile = self.profile_manager.current_profile
        cmd_id = self.profile_manager.cmd_info['cid']

        if profile is None:
            console.warning("Get stop profile commands, but threading profiling is not running. if this continues, "
                            "please report to us for more investigation.")
        elif profile_id != profile.profile_id:
            self._active_session.send_profile_data({"id": cmd_id, "result": {}})
            console.warning("Get stop profile commands, but the specified profile id[%s] not match current"
                            "profile id[%s]. This command will be ignored.", profile_id, profile.profile_id)
            return

        self.profile_manager.shutdown(stop_cmd=True)
        self.profile_status = None
        console.info("Stopping profile for application %s", self._app_name)

    def send_profile_data(self, audit_mode=False):
        """get the profile data and send the agent server
        :return:
        """
        # profile thread start failed. skip report process
        if not self.profile_status:
            return

        profile_data = self.profile_manager.generate_profile_data()

        # dose not finished.
        if profile_data is None:
            return

        if audit_mode:
            console.info("Agent capture the profile data %s", profile_data)

        ret = self._active_session.send_profile_data(profile_data)
        self.profile_status = None
        console.info("send profile return result %s", ret)



import logging

""" define some exception class for internal usage
"""

console = logging.getLogger(__name__)


class ConfigurationError(Exception):
    pass


class NetworkInterfaceException(Exception):
    pass


class ForceAgentRestart(NetworkInterfaceException):
    pass


class ForceAgentDisconnect(NetworkInterfaceException):
    pass


class DiscardDataForRequest(NetworkInterfaceException):
    pass


class RetryDataForRequest(NetworkInterfaceException):
    pass


class ServerIsUnavailable(RetryDataForRequest):
    pass


class InvalidLicenseException(NetworkInterfaceException):
    pass


class InvalidDataTokenException(NetworkInterfaceException):
    pass


class OutOfDateConfigException(NetworkInterfaceException):
    pass


class CommandlineParametersException(Exception):
    pass


# -*- coding: utf-8 -*-

""" define some detector for tracer entrance
"""

import time
import logging

from tingyun.armoury.ammunition.tornado_4.utils import is_websocket, generate_tracer
from tingyun.armoury.ammunition.tornado_4.wrappers import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado_4.name_parser import object_name

console = logging.getLogger(__name__)


def trace_server_request_init_(wrapped, request, args, kwargs):
    """agent tracer entrance in tornado, will be called  after header but body. Three situations will be called.
        1. a common web.Application instance.
        2.instance passed into httpserver.HTTPServer is a not tornado application.
        3. In wsgi module. some WSGI applications.

    :param request: a `request` Instance of current user request.
    """
    result = wrapped(*args, **kwargs)

    tracker = None if is_websocket(request) else generate_tracer(request)
    if tracker:
        tracker.set_tracker_name(object_name(wrapped), priority=1)
        tracker.start_time = getattr(request, "_start_time", int(time.time()))

    request._nb_tracker = tracker
    return result


def detect_tracker_entrance(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "HTTPServerRequest.__init__", trace_server_request_init_)
